# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import io
import os

import pytest

from ovirt_imageio._internal import directio
from ovirt_imageio._internal import errors
from ovirt_imageio._internal import util

from . import testutil

# Legacy code supports only 512 bytes.
BLOCKSIZE = 512


def fill(b, size):
    count, rest = divmod(size, len(b))
    return b * count + b[:rest]


BUFFER = fill(b"ABCDEFGHIJ", 1024**2)
BLOCK = fill(b"abcdefghij", 512)
BYTES = fill(b"0123456789", 42)


@pytest.mark.parametrize("offset", [0, 42, 512])
@pytest.mark.parametrize("data", [
    BUFFER * 2,
    BUFFER + BLOCK * 2,
    BUFFER + BLOCK + BYTES,
    BLOCK * 2,
    BLOCK + BYTES,
    BYTES,
], ids=testutil.head)
def test_receive(tmpfile, data, offset):
    assert receive(tmpfile, data, len(data), offset=offset) == data


@pytest.mark.parametrize("offset", [0, 42, 512])
@pytest.mark.parametrize("size", [
    511,
    513,
    len(BUFFER) + 511,
    len(BUFFER) + 513,
])
def test_receive_partial(tmpfile, size, offset):
    data = BUFFER * 2
    assert receive(tmpfile, data, size, offset=offset) == data[:size]


@pytest.mark.parametrize("offset", [0, 42, 512])
@pytest.mark.parametrize("data", [
    BUFFER * 2,
    BUFFER + BLOCK * 2,
    BUFFER + BLOCK + BYTES,
    BLOCK * 2,
    BLOCK + BYTES,
    BYTES,
], ids=testutil.head)
def test_receive_partial_content(tmpfile, data, offset):
    with pytest.raises(errors.PartialContent) as e:
        receive(tmpfile, data[:-1], len(data), offset=offset)
    assert e.value.requested == len(data)
    assert e.value.available == len(data) - 1


def receive(tmpfile, data, size, offset=0):
    with open(tmpfile, "wb") as f:
        f.write(b"x" * offset)
    src = io.BytesIO(data)
    op = directio.Receive(tmpfile, src, size, offset=offset)
    op.run()
    with open(tmpfile, "rb") as f:
        f.seek(offset)
        return f.read(size)


@pytest.mark.parametrize("extra, calls", [
    ({}, 1),  # Flushes by default.
    ({"flush": True}, 1),
    ({"flush": False}, 0),
])
def test_receive_flush(tmpfile, monkeypatch, extra, calls):
    # This would be much cleaner when we add backend object implementing flush.
    fsync = os.fsync
    fsync_calls = [0]

    def counted_fsync(fd):
        fsync_calls[0] += 1
        fsync(fd)

    monkeypatch.setattr("os.fsync", counted_fsync)
    data = b"x" * 1024**2 * 2
    with open(tmpfile, "wb") as f:
        f.write(data)
    size = len(data)
    src = io.BytesIO(b"X" * size)
    op = directio.Receive(tmpfile, src, size, **extra)
    op.run()
    with io.open(tmpfile, "rb") as f:
        assert f.read() == src.getvalue()
    assert fsync_calls[0] == calls


def test_recv_repr(tmpfile):
    op = directio.Receive(tmpfile, None, 100, offset=42)
    rep = repr(op)
    assert "Receive" in rep
    assert "size=100 offset=42 done=0" in rep


@pytest.mark.parametrize("bufsize", [512, 1024, 2048])
def test_receive_unbuffered_stream(tmpfile, bufsize):
    chunks = [b"a" * 8192,
              b"b" * 42,
              b"c" * (8192 - 42)]
    data = b''.join(chunks)
    assert receive_unbuffered(tmpfile, chunks, len(data), bufsize) == data


def test_receive_unbuffered_stream_partial_content(tmpfile):
    chunks = [b"a" * 8192,
              b"b" * 42,
              b"c" * (8192 - 42)]
    data = b''.join(chunks)
    with pytest.raises(errors.PartialContent):
        receive_unbuffered(tmpfile, chunks, len(data) + 1, 2048)


def receive_unbuffered(tmpfile, chunks, size, bufsize):
    src = util.UnbufferedStream(chunks)
    op = directio.Receive(tmpfile, src, size, buffersize=bufsize)
    op.run()
    with open(tmpfile, "rb") as f:
        return f.read()


@pytest.mark.parametrize("offset", [0, 42, 512])
@pytest.mark.parametrize("data", [
    BUFFER * 2,
    BUFFER + BLOCK * 2,
    BUFFER + BLOCK + BYTES,
    BLOCK * 2,
    BLOCK + BYTES,
], ids=testutil.head)
def test_receive_no_size(tmpfile, data, offset):
    with open(tmpfile, "wb") as f:
        f.write(b"x" * offset)
    src = io.BytesIO(data)
    op = directio.Receive(tmpfile, src, offset=offset)
    op.run()
    with io.open(tmpfile, "rb") as f:
        f.seek(offset)
        assert f.read(len(data)) == data


def test_receive_padd_to_block_size(tmpfile):
    with open(tmpfile, "wb") as f:
        f.write(b"x" * 400)
    size = 200
    offset = 300
    padding = BLOCKSIZE - size - offset
    src = io.BytesIO(b"y" * size)
    op = directio.Receive(tmpfile, src, size, offset=offset)
    op.run()
    with open(tmpfile, "rb") as f:
        # Data before offset is not modified.
        assert f.read(300) == b"x" * offset
        # Data after offset is modifed, flie extended.
        assert f.read(200) == b"y" * size
        # File padded to block size with zeroes.
        assert f.read() == b"\0" * padding



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import http.client as http_client
import io
import json
import os
import stat

import pytest

from ovirt_imageio._internal import config
from ovirt_imageio._internal import server

from . import http
from . import testutil


@pytest.fixture(scope="module")
def srv():
    cfg = config.load(["test/conf/daemon.conf"])
    s = server.Server(cfg)
    s.start()
    try:
        yield s
    finally:
        s.stop()


def test_method_not_allowed(srv):
    with http.LocalClient(srv.config) as c:
        res = c.request("FOO", "/images/")
        assert res.status == http_client.METHOD_NOT_ALLOWED


@pytest.mark.parametrize("method", ["GET", "PUT", "PATCH"])
def test_no_resource(srv, method):
    with http.LocalClient(srv.config) as c:
        res = c.request(method, "/no/resource")
        assert res.status == http_client.NOT_FOUND


@pytest.mark.parametrize("method", ["GET", "PUT", "PATCH"])
def test_no_ticket_id(srv, method):
    with http.LocalClient(srv.config) as c:
        res = c.request(method, "/images/")
        assert res.status == http_client.BAD_REQUEST


@pytest.mark.parametrize("method,body", [
    ("GET", None),
    ("PUT", "body"),
    ("PATCH", json.dumps({"op": "flush"}).encode("ascii")),
])
def test_no_ticket(srv, method, body):
    with http.LocalClient(srv.config) as c:
        res = c.request(method, "/images/no-ticket", body=body)
        assert res.status == http_client.FORBIDDEN


def test_put_forbidden(srv):
    ticket = testutil.create_ticket(
        url="file:///no/such/image", ops=["read"])
    srv.auth.add(ticket)
    with http.LocalClient(srv.config) as c:
        res = c.put("/images/" + ticket["uuid"], "content")
        assert res.status == http_client.FORBIDDEN


def test_put(srv, tmpdir):
    data = b"-------|after"
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    uri = "/images/" + ticket["uuid"]
    with http.LocalClient(srv.config) as c:
        res = c.put(uri, "content")
        assert res.status == http_client.OK
        assert res.getheader("content-length") == "0"
        with io.open(str(image)) as f:
            assert f.read(len(data)) == "content|after"


def test_get_forbidden(srv):
    ticket = testutil.create_ticket(
        url="file:///no/such/image", ops=[])
    srv.auth.add(ticket)
    with http.LocalClient(srv.config) as c:
        res = c.get("/images/" + ticket["uuid"], {"header": "content"})
        assert res.status == http_client.FORBIDDEN


def test_get(srv, tmpdir):
    data = b"a" * 512 + b"b" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(
        url="file://" + str(image), size=1024)
    srv.auth.add(ticket)
    with http.LocalClient(srv.config) as c:
        res = c.get("/images/" + ticket["uuid"])
        assert res.status == http_client.OK
        assert res.read() == data


def test_images_zero(srv, tmpdir):
    data = b"x" * 512
    image = testutil.create_tempfile(tmpdir, "image", data)
    ticket = testutil.create_ticket(url="file://" + str(image))
    srv.auth.add(ticket)
    msg = {"op": "zero", "size": 20, "offset": 10, "future": True}
    size = msg["size"]
    offset = msg.get("offset", 0)
    body = json.dumps(msg).encode("ascii")
    with http.LocalClient(srv.config) as c:
        res = c.patch("/images/" + ticket["uuid"], body)
        assert res.status == http_client.OK
        assert res.getheader("content-length") == "0"
        with io.open(str(image), "rb") as f:
            assert f.read(offset) == data[:offset]
            assert f.read(size) == b"\0" * size
            assert f.read() == data[offset + size:]


def test_options(srv):
    with http.LocalClient(srv.config) as c:
        res = c.options("/images/*")
        allows = {"OPTIONS", "GET", "PUT", "PATCH"}
        features = {"checksum", "extents", "flush", "zero"}
        assert res.status == http_client.OK
        assert set(res.getheader("allow").split(',')) == allows
        options = json.loads(res.read())
        assert set(options["features"]) == features
        assert options["unix_socket"] == srv.local_service.address


def test_control_socket_mode(srv):
    socket = srv.config.control.socket
    actual_mode = stat.S_IMODE(os.stat(socket).st_mode)
    assert oct(actual_mode) == oct(0o660)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Manage imageio server.

This module exposes the public names. Anything else in this package is private
and should not be used.
"""

# flake8: noqa

from .. _internal import version

from .. _internal.server import (
    DEFAULT_CONF_DIR,
    load_config,
)

from . _api import (
    Client,
    ClientError,
    Error,
    ServerError,
)

__all__ = (
    "Client",
    "ClientError",
    "DEFAULT_CONF_DIR"
    "Error",
    "ServerError",
    "load_config",
)

__version__ = version.string



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import logging

from .. import backends
from .. import errors
from .. import http
from .. import validate

log = logging.getLogger("extents")


class Handler:
    """
    Handle requests for the /images/ticket-id/extents resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    def get(self, req, resp, ticket_id):
        if not ticket_id:
            raise http.Error(http.BAD_REQUEST, "Ticket id is required")

        try:
            ticket = self.auth.authorize(ticket_id, "read")
            ctx = backends.get(req, ticket, self.config)
        except errors.AuthorizationError as e:
            resp.close_connection()
            raise http.Error(http.FORBIDDEN, str(e))

        context = validate.enum(
            req.query, "context", ("zero", "dirty"), default="zero")

        if context == "dirty" and not ticket.dirty:
            raise http.Error(
                http.NOT_FOUND, "Ticket does not support dirty extents")

        log.info("[%s] EXTENTS transfer=%s context=%s",
                 req.client_addr, ticket.transfer_id, context)

        with req.clock.run("extents"):
            try:
                extents = [ext.to_dict()
                           for ext in ctx.backend.extents(context=context)]
            except errors.UnsupportedOperation as e:
                raise http.Error(http.NOT_FOUND, str(e))

        resp.send_json(extents)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Manage ovirt-imageio service
"""

import argparse
import json
import sys

from ovirt_imageio import admin


def main():
    parser = argparse.ArgumentParser(
        description="Control the ovirt-imageio service")

    parser.set_defaults(command=None)
    commands = parser.add_subparsers(title="commands")

    add_cmd = add_command(
        commands,
        name="add-ticket",
        help="Add a ticket.",
        command=add_ticket)

    add_cmd.add_argument(
        "filename",
        help="Ticket filename.")

    show_cmd = add_command(
        commands,
        name="show-ticket",
        help="Show a ticket status.",
        command=show_ticket)

    show_cmd.add_argument(
        "ticket_id",
        help="Ticket id.")

    mod_cmd = add_command(
        commands,
        name="mod-ticket",
        help="Modify a ticket.",
        command=mod_ticket)

    mod_cmd.add_argument(
        "ticket_id",
        help="Ticket id.")

    mod_cmd.add_argument(
        "--timeout",
        type=int,
        help="New timeout in seconds.")

    del_cmd = add_command(
        commands,
        name="del-ticket",
        help="Delete a ticket.",
        command=del_ticket)

    del_cmd.add_argument(
        "ticket_id",
        help="Ticket id.")

    add_command(
        commands,
        name="start-profile",
        help="Start server profiling",
        command=start_profile)

    add_command(
        commands,
        name="stop-profile",
        help="Stop server profiling",
        command=stop_profile)

    args = parser.parse_args()
    if not args.command:
        parser.print_help()
        sys.exit(0)

    try:
        args.command(args)
    except admin.Error as e:
        # Expected errror, log a clean error.
        sys.stderr.write(f"ovirt-imageioctl: {e}\n")
        sys.exit(1)


def add_command(commands, name, help, command):
    cmd = commands.add_parser(name, help=help)
    cmd.set_defaults(command=command)
    cmd.add_argument(
        "-c", "--conf-dir",
        default=admin.DEFAULT_CONF_DIR,
        help=f"Configuration directory (default {admin.DEFAULT_CONF_DIR}).")
    return cmd


def add_ticket(args):
    with open(args.filename) as f:
        data = f.read()
    ticket = json.loads(data)
    cfg = admin.load_config(args.conf_dir)
    with admin.Client(cfg) as c:
        c.add_ticket(ticket)


def show_ticket(args):
    cfg = admin.load_config(args.conf_dir)
    with admin.Client(cfg) as c:
        info = c.get_ticket(args.ticket_id)
    print(json.dumps(info, indent=2))


def mod_ticket(args):
    cfg = admin.load_config(args.conf_dir)
    with admin.Client(cfg) as c:
        c.mod_ticket(args.ticket_id, {"timeout": args.timeout})


def del_ticket(args):
    cfg = admin.load_config(args.conf_dir)
    with admin.Client(cfg) as c:
        c.del_ticket(args.ticket_id)


def start_profile(args):
    cfg = admin.load_config(args.conf_dir)
    with admin.Client(cfg) as c:
        c.start_profile()


def stop_profile(args):
    cfg = admin.load_config(args.conf_dir)
    with admin.Client(cfg) as c:
        c.stop_profile()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Download commands.
"""

from contextlib import closing

from . import _api
from . import _options
from . import _ovirt
from . import _ui


def register(parser):
    cmd = parser.add_sub_command(
        "download-disk",
        help="Download disk",
        func=download_disk)

    cmd.add_argument(
        "-f", "--format",
        choices=("raw", "qcow2"),
        default="qcow2",
        help="Download image format (default qcow2).")

    cmd.add_argument(
        "disk_id",
        type=_options.UUID,
        help="Disk ID to download.")

    cmd.add_argument(
        "filename",
        help="Target filename.")


def download_disk(args):
    with _ui.ProgressBar(phase="creating transfer", format=args.output) as pb:
        con = _ovirt.connect(args)
        with closing(con):
            disk = _ovirt.find_disk(con, args.disk_id)
            storage_domain = _ovirt.find_storage_domain(con, disk)
            host = _ovirt.find_host(con, storage_domain.name)

            transfer = _ovirt.create_transfer(
                con, disk, direction=_ovirt.DOWNLOAD, host=host)
            try:
                pb.phase = "downloading image"
                _api.download(
                    transfer.transfer_url,
                    args.filename,
                    args.cafile,
                    fmt=args.format,
                    secure=args.secure,
                    proxy_url=transfer.proxy_url,
                    max_workers=args.max_workers,
                    buffer_size=args.buffer_size,
                    progress=pb)
            finally:
                pb.phase = "finalizing transfer"
                _ovirt.finalize_transfer(con, transfer, disk)
        pb.phase = "download completed"



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Upload commands.
"""

import os
from contextlib import closing
from collections import namedtuple

from . import _api
from . import _options
from . import _ovirt
from . import _ui


DiskInfo = namedtuple(
    "DiskInfo",
    "name,initial_size,provisioned_size,content_type,format,sparse,is_zero")

FORMAT_RAW = "raw"
FORMAT_QCOW2 = "qcow2"
_DISK_FORMATS = (FORMAT_RAW, FORMAT_QCOW2)


def register(parser):
    cmd = parser.add_sub_command(
        "upload-disk",
        help="Upload disk",
        func=upload_disk)

    cmd.add_argument(
        "-s", "--storage-domain",
        required=True,
        help="Name of the storage domain.")

    cmd.add_argument(
        "-f", "--format",
        choices=_DISK_FORMATS,
        default=FORMAT_QCOW2,
        help="Upload disk format (default qcow2 for data disks and raw "
             "for iso disks).")

    cmd.add_argument(
        "--preallocated",
        dest="sparse",
        action="store_false",
        help="Create preallocated disk. Required when using raw format on "
             "block based storage domain (iSCSI, FC). ISO images are "
             "always uploaded to preallocated disk.")

    cmd.add_argument(
        "--disk-id",
        type=_options.UUID,
        help="A UUID for the new disk. If not specified, oVirt will "
             "create a new UUID.")

    cmd.add_argument(
        "--name",
        help="Alias name for the new disk. If not specified, name will "
             "correspond with the image filename.")

    cmd.add_argument(
        "filename",
        type=_options.File,
        help="Path to image to upload. Supported formats: raw, qcow2, iso.")


def upload_disk(args):
    with _ui.ProgressBar(
            phase="inspecting image", format=args.output) as progress:
        disk_info = _prepare(args)
        con = _ovirt.connect(args)
        with closing(con):
            progress.phase = "creating disk"
            disk = _ovirt.add_disk(
                con=con,
                name=disk_info.name,
                initial_size=disk_info.initial_size,
                provisioned_size=disk_info.provisioned_size,
                sd_name=args.storage_domain,
                id=args.disk_id,
                sparse=disk_info.sparse,
                enable_backup=disk_info.format == _ovirt.COW,
                content_type=disk_info.content_type,
                format=disk_info.format,
                timeout=args.disk_timeout)

            progress.phase = "creating transfer"
            host = _ovirt.find_host(con, args.storage_domain)
            transfer = _ovirt.create_transfer(con, disk, host=host)
            try:
                progress.phase = "uploading image"
                _api.upload(
                    args.filename,
                    transfer.transfer_url,
                    args.cafile,
                    buffer_size=args.buffer_size,
                    progress=progress,
                    secure=args.secure,
                    proxy_url=transfer.proxy_url,
                    max_workers=args.max_workers,
                    disk_is_zero=disk_info.is_zero)
            except Exception:
                progress.phase = "cancelling transfer"
                _ovirt.cancel_transfer(con, transfer)
                raise

            progress.phase = "finalizing transfer"
            _ovirt.finalize_transfer(con, transfer, disk)
        progress.phase = "upload completed"


def _prepare(args):
    # Obtain the image info dictionary.
    img_info = _api.info(args.filename)
    if img_info["format"] not in _DISK_FORMATS:
        raise RuntimeError(f"Unsupported image format {img_info['format']}")

    # Obtain all remaining fields for DiskInfo.
    if _is_iso(img_info["filename"], img_info["format"]):
        content_type = _ovirt.ISO
        # ISO images require raw format to work with cdrom device
        disk_format = FORMAT_RAW
        sparse = False
    else:
        content_type = _ovirt.DATA
        disk_format = args.format
        sparse = args.sparse

    initial_size = None
    if disk_format == FORMAT_QCOW2 and sparse:
        initial_size = _api.measure(args.filename, disk_format)["required"]

    name = args.name
    if name is None:
        name = os.path.splitext(os.path.basename(img_info["filename"]))[0]

    # On file based storage new disk is always zero. On block based storage
    # only when using sparse qcow2 format. Since we don't know the type of the
    # storage domain, we cannot optimize all cases.
    is_zero = (disk_format == FORMAT_QCOW2) or sparse

    return DiskInfo(
        name=name,
        initial_size=initial_size,
        provisioned_size=img_info["virtual-size"],
        content_type=content_type,
        format=_ovirt.COW if disk_format == FORMAT_QCOW2 else _ovirt.RAW,
        sparse=sparse,
        is_zero=is_zero)


def _is_iso(filename, image_format):
    """
    Detect if disk content type is ISO

    ISO format structure
    ---------------------------------------------------------------------------
    offset    type    value       comment
    ---------------------------------------------------------------------------
    0x0000                        system area (e.g. DOS/MBR boot sector)
    0x8000    int8    0x01        primary volume descriptor type code
    0x8001    strA    "CD001"     primary volume descriptor indentifier
    0x8006    int8    0x01        primary volume desctptor version
    0x8007            0x00        unused field

    See https://wiki.osdev.org/ISO_9660#Overview_and_caveats for more info.

    Returns:
        bool: Source image is ISO
    """
    if image_format == FORMAT_RAW:
        with open(filename, "rb") as file:
            file.seek(0x8000)
            primary_volume_descriptor = file.read(8)
        if primary_volume_descriptor == b"\x01CD001\x01\x00":
            return True
    return False



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import hashlib
import pytest
from functools import partial

from ovirt_imageio._internal import blkhash

default_hash = partial(hashlib.blake2b, digest_size=32)


def test_algorithm_basic():
    h = default_hash()
    for i in range(10):
        block = (b"%02d\n" % i).ljust(blkhash.BLOCK_SIZE, b"\0")
        block_digest = default_hash(block).digest()
        h.update(block_digest)
    assert h.hexdigest() == (
        "7934079f80b53142d738d2bb7efaedf696a3d34d76a7865a24130bc7b4a7acfe"
    )


def test_algorithm_zero_optimization():
    # Hash the entire image. This is the case when we don't have extent
    # information, and do not use zero detection.
    zero_block = b"\0" * blkhash.BLOCK_SIZE
    h1 = default_hash()
    for i in range(10):
        block_digest = default_hash(zero_block).digest()
        h1.update(block_digest)

    # Hash a pre-computed digest instead of the actual bytes. Here we either
    # have extent information, or we detected that the blocks are zero blocks.
    h2 = default_hash()
    block_digest = default_hash(zero_block).digest()
    for i in range(10):
        h2.update(block_digest)

    # We must get the same checksum in both cases.
    assert h1.hexdigest() == h2.hexdigest()


def test_hasher_data():
    h1 = blkhash.Hash()
    for i in range(10):
        block = (b"%02d\n" % i).ljust(blkhash.BLOCK_SIZE, b"\0")
        h1.update(block)

    h2 = default_hash()
    for i in range(10):
        block = (b"%02d\n" % i).ljust(blkhash.BLOCK_SIZE, b"\0")
        block_digest = default_hash(block).digest()
        h2.update(block_digest)

    assert h1.hexdigest() == h2.hexdigest()


def test_hasher_zero():
    block = b"\0" * blkhash.BLOCK_SIZE

    h1 = blkhash.Hash()
    h1.update(block)
    h1.update(block)

    h2 = blkhash.Hash()
    h2.zero(len(block))
    h2.zero(len(block))

    assert h1.hexdigest() == h2.hexdigest()


@pytest.mark.parametrize("size,algorithm,digest_size,checksum", [
    # Files aligned to block size.
    (4 * 1024**2, "blake2b", 32,
        "f426bb2cf1e1901fe4e87423950944ecfed6d9d18a09e6e802aa4912e1c9b2d6"),
    (4 * 1024**2, "sha1", None,
        "3ed03b375b6658d99b63ced1867a95aeef080b79"),
    # Files not aligned to block size.
    (3 * 1024**2, "blake2b", 32,
        "42f3b76772a6d3dcffae2a24697721687975e2c60ddfd4ba7831ea9ce772ca71"),
    (3 * 1024**2, "sha1", None,
        "6cba43b908381be45a55ab9b4361f8370b928354"),
    (5 * 1024**2, "blake2b", 32,
        "0da53b583fc1fbbac7edea14454c79f84a8107613e614f2c7a47071dfdcf41a6"),
    (5 * 1024**2, "sha1", None,
        "d3936edd8e3a8ff10e8257a9f460d8da67838549"),
    # Empty file.
    (0, "blake2b", 32,
        "0e5751c026e543b2e8ab2eb06099daa1d1e5df47778f7787faab45cdf12fe3a8"),
    (0, "sha1", None,
        "da39a3ee5e6b4b0d3255bfef95601890afd80709"),
])
def test_checksum(tmpdir, size, algorithm, digest_size, checksum):
    path = str(tmpdir.join("file"))

    with open(path, "wb") as f:
        f.write(b"data")
        f.truncate(size)

    actual = blkhash.checksum(
        path,
        block_size=blkhash.BLOCK_SIZE,
        algorithm=algorithm,
        digest_size=digest_size)

    assert actual == {
        "algorithm": algorithm,
        "block_size": blkhash.BLOCK_SIZE,
        "checksum": checksum,
    }



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import os

from contextlib import contextmanager

import pytest

from ovirt_imageio._internal import auth
from ovirt_imageio._internal import config
from ovirt_imageio._internal import services
from ovirt_imageio._internal.ssl import check_protocol


@contextmanager
def remote_service(config_file):
    path = os.path.join("test/conf", config_file)
    cfg = config.load([path])
    authorizer = auth.Authorizer(cfg)
    s = services.RemoteService(cfg, authorizer)
    s.start()
    try:
        yield s
    finally:
        s.stop()


@pytest.mark.parametrize("protocol", ["-ssl2", "-ssl3", "-tls1", "-tls1_1"])
def test_default_reject(protocol):
    with remote_service("daemon.conf") as service:
        rc = check_protocol("127.0.0.1", service.port, protocol)
    assert rc != 0


@pytest.mark.parametrize("protocol", ["-tls1_2", "-tls1_3"])
def test_default_accept(protocol):
    with remote_service("daemon.conf") as service:
        rc = check_protocol("127.0.0.1", service.port, protocol)
    assert rc == 0



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import os
import subprocess

from contextlib import closing

import pytest

from ovirt_imageio._internal import ioutil
from ovirt_imageio._internal import util

BLOCKSIZE = 4096

requires_root = pytest.mark.skipif(os.geteuid() != 0, reason="Requires root")


@pytest.fixture
def loop_device(tmpdir):
    backing_file = str(tmpdir.join("backing_file"))
    with util.open(backing_file, "w") as f:
        buf = util.aligned_buffer(BLOCKSIZE * 3)
        with closing(buf):
            buf[:] = b"x" * BLOCKSIZE * 3
            f.write(buf)
    out = subprocess.check_output(
        ["losetup", "--find", backing_file, "--show"])
    try:
        loop = out.strip().decode("ascii")
        yield loop
    finally:
        subprocess.check_call(["losetup", "--detach", loop])


@requires_root
def test_zeroout_start(loop_device):
    with util.open(loop_device, "r+") as f:
        ioutil.blkzeroout(f.fileno(), 0, BLOCKSIZE)

    with util.open(loop_device, "r") as f:
        buf = util.aligned_buffer(BLOCKSIZE * 3)
        with closing(buf):
            f.readinto(buf)
            assert buf[:BLOCKSIZE] == b"\0" * BLOCKSIZE
            assert buf[BLOCKSIZE:] == b"x" * BLOCKSIZE * 2


@requires_root
def test_zeroout_middle(loop_device):
    with util.open(loop_device, "r+") as f:
        ioutil.blkzeroout(f.fileno(), BLOCKSIZE, BLOCKSIZE)

    with util.open(loop_device, "r") as f:
        buf = util.aligned_buffer(BLOCKSIZE * 3)
        with closing(buf):
            f.readinto(buf)
            assert buf[:BLOCKSIZE] == b"x" * BLOCKSIZE
            assert buf[BLOCKSIZE:-BLOCKSIZE] == b"\0" * BLOCKSIZE
            assert buf[-BLOCKSIZE:] == b"x" * BLOCKSIZE


@requires_root
def test_zeroout_end(loop_device):
    with util.open(loop_device, "r+") as f:
        ioutil.blkzeroout(f.fileno(), BLOCKSIZE * 2, BLOCKSIZE)

    with util.open(loop_device, "r") as f:
        buf = util.aligned_buffer(BLOCKSIZE * 3)
        with closing(buf):
            f.readinto(buf)
            assert buf[:-BLOCKSIZE] == b"x" * BLOCKSIZE * 2
            assert buf[-BLOCKSIZE:] == b"\0" * BLOCKSIZE


@requires_root
def test_blksszget_512(loop_device):
    with util.open(loop_device, "r+") as f:
        assert ioutil.blksszget(f.fileno()) == 512


@requires_root
def test_blksszget_bad_fd(loop_device):
    with pytest.raises(OSError):
        ioutil.blksszget(-1)


@requires_root
def test_blksszget_not_block_device(loop_device, tmpfile):
    with open(tmpfile) as f:
        with pytest.raises(OSError):
            ioutil.blksszget(f.fileno())


# Empty zero buffer

@pytest.mark.parametrize("buf", [
    pytest.param("", id="unicode"),
    pytest.param(b"", id="bytes"),
    pytest.param(bytearray(), id="bytearray"),
])
def test_is_zero_empty(buf):
    assert ioutil.is_zero(buf)


@pytest.mark.parametrize("buf", [
    pytest.param(b"", id="bytes"),
    pytest.param(bytearray(), id="bytearray"),
])
def test_is_zero_empty_memoryview(buf):
    assert ioutil.is_zero(memoryview(buf))


# Non-empty zero buffer

@pytest.mark.parametrize("buf", [
    pytest.param("\u0000" * 512, id="unicode"),
    pytest.param(b"\0" * 512, id="bytes"),
    pytest.param(bytearray(512), id="bytearray"),
])
def test_is_zero(buf):
    assert ioutil.is_zero(buf)


@pytest.mark.parametrize("buf", [
    pytest.param(b"\0" * 512, id="bytes"),
    pytest.param(bytearray(512), id="bytearray"),
])
def test_is_zero_memoryview(buf):
    assert ioutil.is_zero(memoryview(buf))


# Non-zero buffer with non-zero in first 16 bytes

@pytest.mark.parametrize("buf", [
    pytest.param("\u0000" * 15 + "x", id="unicode"),
    pytest.param(b"\0" * 15 + b"x", id="bytes"),
    pytest.param(bytearray(b"\0" * 15 + b"x"), id="bytearray"),
])
def test_is_not_zero_head(buf):
    assert not ioutil.is_zero(buf)


@pytest.mark.parametrize("buf", [
    pytest.param(b"\0" * 15 + b"x", id="bytes"),
    pytest.param(bytearray(b"\0" * 15 + b"x"), id="bytearray"),
])
def test_is_not_zero_head_memoryview(buf):
    assert not ioutil.is_zero(memoryview(buf))


# Non-zero buffer

@pytest.mark.parametrize("buf", [
    pytest.param("\u0000" * 511 + "x", id="unicode"),
    pytest.param(b"\0" * 511 + b"x", id="bytes"),
    pytest.param(bytearray(b"\0" * 511 + b"x"), id="bytearray"),
])
def test_is_not_zero(buf):
    assert not ioutil.is_zero(buf)


@pytest.mark.parametrize("buf", [
    pytest.param(b"\0" * 511 + b"x", id="bytes"),
    pytest.param(bytearray(b"\0" * 511 + b"x"), id="bytearray"),
])
def test_is_not_zero_memoryview(buf):
    assert not ioutil.is_zero(memoryview(buf))


# Checking mmap

@pytest.fixture
def aligned_buffer():
    buf = util.aligned_buffer(BLOCKSIZE)
    with closing(buf):
        yield buf


def test_is_zero_mmap(aligned_buffer):
    assert ioutil.is_zero(aligned_buffer)


def test_is_not_zero_mmap_head(aligned_buffer):
    aligned_buffer[15:16] = b"x"
    assert not ioutil.is_zero(aligned_buffer)


def test_is_not_zero_mmap(aligned_buffer):
    aligned_buffer[-1:] = b"x"
    assert not ioutil.is_zero(aligned_buffer)


# fallocate

fallocate_mode = pytest.mark.parametrize("mode", [
    # Zero byte range, allocating space - for preallocated images.
    # Not supported on NFS 4.2.
    ioutil.FALLOC_FL_ZERO_RANGE,
    # Zero byte range, daallocating space - for sparse images.
    # Supported on NFS 4.2.
    ioutil.FALLOC_FL_PUNCH_HOLE | ioutil.FALLOC_FL_KEEP_SIZE,
])


@fallocate_mode
def test_fallocate_zero_start(tmpdir, mode):
    path = str(tmpdir.join("file"))
    with open(path, "wb") as f:
        f.write(b"x" * BLOCKSIZE * 3)

    buf = util.aligned_buffer(BLOCKSIZE * 3)
    with closing(buf), util.open(path, "r+") as f:
        try_fallocate(f.fileno(), mode, 0, BLOCKSIZE)

        n = f.readinto(buf)
        assert n == BLOCKSIZE * 3
        assert buf[:BLOCKSIZE] == b"\0" * BLOCKSIZE
        assert buf[BLOCKSIZE:] == b"x" * BLOCKSIZE * 2
        assert f.readinto(buf) == 0


@fallocate_mode
def test_fallocate_zero_middle(tmpdir, mode):
    path = str(tmpdir.join("file"))
    with open(path, "wb") as f:
        f.write(b"x" * BLOCKSIZE * 3)

    buf = util.aligned_buffer(BLOCKSIZE * 3)
    with closing(buf), util.open(path, "r+") as f:
        try_fallocate(f.fileno(), mode, BLOCKSIZE, BLOCKSIZE)

        n = f.readinto(buf)
        assert n == BLOCKSIZE * 3
        assert buf[:BLOCKSIZE] == b"x" * BLOCKSIZE
        assert buf[BLOCKSIZE:-BLOCKSIZE] == b"\0" * BLOCKSIZE
        assert buf[-BLOCKSIZE:] == b"x" * BLOCKSIZE
        assert f.readinto(buf) == 0


@fallocate_mode
def test_fallocate_zero_end(tmpdir, mode):
    path = str(tmpdir.join("file"))
    with open(path, "wb") as f:
        f.write(b"x" * BLOCKSIZE * 3)

    buf = util.aligned_buffer(BLOCKSIZE * 3)
    with closing(buf), util.open(path, "r+") as f:
        try_fallocate(f.fileno(), mode, BLOCKSIZE * 2, BLOCKSIZE)

        n = f.readinto(buf)
        assert n == BLOCKSIZE * 3
        assert buf[:-BLOCKSIZE] == b"x" * BLOCKSIZE * 2
        assert buf[-BLOCKSIZE:] == b"\0" * BLOCKSIZE
        assert f.readinto(buf) == 0


def test_fallocate_zero_after_end(tmpdir):
    path = str(tmpdir.join("file"))
    with open(path, "wb") as f:
        f.write(b"x" * BLOCKSIZE * 3)

    buf = util.aligned_buffer(BLOCKSIZE * 4)
    with closing(buf), util.open(path, "r+") as f:
        # Will allocate more space that will return zeros when read.
        mode = ioutil.FALLOC_FL_ZERO_RANGE
        try_fallocate(f.fileno(), mode, BLOCKSIZE * 3, BLOCKSIZE)

        n = f.readinto(buf)
        assert n == BLOCKSIZE * 4
        assert buf[:-BLOCKSIZE] == b"x" * BLOCKSIZE * 3
        assert buf[-BLOCKSIZE:] == b"\0" * BLOCKSIZE
        assert f.readinto(buf) == 0


def test_fallocate_punch_hole_after_end(tmpdir):
    path = str(tmpdir.join("file"))
    with open(path, "wb") as f:
        f.write(b"x" * BLOCKSIZE * 3)

    buf = util.aligned_buffer(BLOCKSIZE * 3)
    with closing(buf), util.open(path, "r+") as f:
        # This does not change file contents or size.
        mode = ioutil.FALLOC_FL_PUNCH_HOLE | ioutil.FALLOC_FL_KEEP_SIZE
        try_fallocate(f.fileno(), mode, BLOCKSIZE * 3, BLOCKSIZE)

        n = f.readinto(buf)
        assert n == BLOCKSIZE * 3
        assert buf[:] == b"x" * BLOCKSIZE * 3
        assert f.readinto(buf) == 0


def try_fallocate(fd, mode, offset, count):
    try:
        ioutil.fallocate(fd, mode, offset, count)
    except EnvironmentError as e:
        if e.errno != errno.EOPNOTSUPP:
            raise
        pytest.skip("fallocate(mode=%r) not supported" % mode)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import collections
import io
import mmap
import os
import shutil
import tempfile
import threading

from contextlib import contextmanager

from .units import KiB


def start_thread(func, args=(), kwargs=None, name=None, daemon=True):
    if kwargs is None:
        kwargs = {}
    t = threading.Thread(target=func, args=args, kwargs=kwargs, name=name)
    t.daemon = daemon
    t.start()
    return t


def monotonic_time():
    return os.times()[4]


def humansize(n):
    for unit in ("bytes", "KiB", "MiB", "GiB", "TiB", "PiB"):
        if n < KiB:
            break
        n /= KiB
    return "{:.{precision}f} {}".format(
        n, unit, precision=0 if unit == "bytes" else 2)


def round_up(n, size):
    n = n + size - 1
    return n - (n % size)


def round_down(n, size):
    return n - (n % size)


def aligned_buffer(size):
    """
    Return buffer aligned to page size, which work for doing direct I/O.

    Note: we use shared map to make direct io safe if fork is invoked in
    another thread concurrently with the direct io.

    Using private maps with direct io can cause data corruption and undefined
    behavior in the parent or the child processes. This restriction does not
    apply to memory buffer created with MAP_SHARED. See open(2) for more info.
    """
    return mmap.mmap(-1, size, mmap.MAP_SHARED)


def open(path, mode, direct=True, sync=False):
    """
    Open a file for direct I/O.

    Writing or reading from the file requires an aligned buffer. Only
    readinto() can be used to read from the file.

    Arguments:
        path (str): Filesystem path
        mode (str): One of ("r", "w", "r+"). The file is always opened in
            binary mode. See io.FileIO for more info on available modes.
        direct (bool): Try to minimize cache effects of the I/O to and from
            this file (O_DIRECT).
        sync (bool): Write operations on the file will complete according to
            the requirements of synchronized I/O file integrity completion
            (O_SYNC).
    """
    if mode == "r":
        flags = os.O_RDONLY
    elif mode == "w":
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
    elif mode == "r+":
        flags = os.O_RDWR
    else:
        raise ValueError("Unsupported mode %r" % mode)

    if direct:
        flags |= os.O_DIRECT

    if sync:
        flags |= os.O_SYNC

    fd = os.open(path, flags)
    return io.FileIO(fd, mode, closefd=True)


def ensure_text(s, encoding='utf-8', errors='strict'):
    """
    Converts:

      - `str` -> `str`
      - `bytes` -> decoded to `str`

    Based on implementation in six library.
    """
    if isinstance(s, bytes):
        return s.decode(encoding, errors)
    elif isinstance(s, str):
        return s
    else:
        raise TypeError("not expecting type '%s'" % type(s))


@contextmanager
def tmp_dir(prefix):
    path = tempfile.mkdtemp(prefix=prefix)
    try:
        yield path
    finally:
        shutil.rmtree(path)


class UnbufferedStream:
    """
    Unlike regular file object, read may return any amount of bytes up to the
    requested size. This behavior is probably the result of doing one syscall
    per read, without any buffering.

    This stream will break code assuming that read(n) retruns n bytes. This
    assumption is normally true, but not all file-like objects behave in this
    way.

    This simulate libvirt stream behavior used to copy imaged directly from
    libvirt.
    https://libvirt.org/html/libvirt-libvirt-stream.html#virStreamRecv
    """

    def __init__(self, chunks):
        self.chunks = collections.deque(chunks)

    def read(self, size):
        if not self.chunks:
            return b''
        chunk = self.chunks.popleft()
        res = chunk[:size]
        chunk = chunk[size:]
        if chunk:
            self.chunks.appendleft(chunk)
        return res

    def readinto(self, buf):
        chunk = self.read(len(buf))
        buf[:len(chunk)] = chunk
        return len(chunk)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import os
import pprint
import tempfile

from contextlib import closing
from contextlib import contextmanager

import pytest

from ovirt_imageio._internal import util
from ovirt_imageio._internal import uhttp


@pytest.fixture(scope="session")
def uhttpserver(request):
    tmp = tempfile.NamedTemporaryFile()
    server = uhttp.Server(tmp.name, uhttp.Connection)
    util.start_thread(server.serve_forever, kwargs={"poll_interval": 0.1})
    request.addfinalizer(server.shutdown)
    request.addfinalizer(tmp.close)
    return server


def test_get(uhttpserver):
    uhttpserver.app = get
    with make_connection(uhttpserver) as con:
        con.request("GET", "/")
        resp = con.getresponse()
        log_response(resp)
        assert resp.status == 200
        assert resp.getheader("content-type") == "text/plain"
        assert resp.read() == b"it works"


def test_put(uhttpserver):
    uhttpserver.app = echo
    with make_connection(uhttpserver) as con:
        con.request("PUT", "/", body=b"it works")
        resp = con.getresponse()
        log_response(resp)
        assert resp.status == 200
        assert resp.getheader("content-type") == "text/plain"
        assert resp.read() == b"it works"


def test_file(tmpdir, uhttpserver):
    data = b"x" * 1048576
    tmp = tmpdir.join("data")
    tmp.write(data)
    uhttpserver.app = sendfile
    with make_connection(uhttpserver) as con:
        con.request("GET", str(tmp))
        resp = con.getresponse()
        log_response(resp)
        assert resp.status == 200
        assert resp.getheader("content-type") == "text/plain"
        content_length = int(resp.getheader("content-length"))
        assert content_length == os.path.getsize(str(tmp))
        assert resp.read(content_length) == data


def test_connection_set_tunnel(uhttpserver):
    with make_connection(uhttpserver) as con:
        with pytest.raises(uhttp.UnsupportedError):
            con.set_tunnel("127.0.0.1")


@pytest.mark.skipif(os.geteuid() == 0,
                    reason="Not compatible when running with root")
def test_server_bind_error(tmpdir):
    # Make server_bind fail with EPERM
    tmpdir.chmod(0o600)
    try:
        sock = str(tmpdir.join('sock'))
        with pytest.raises(OSError):
            uhttp.Server(sock, uhttp.Connection)
    finally:
        tmpdir.chmod(0o755)


def get(req, resp):
    body = b"it works"
    resp.headers["content-length"] = len(body)
    resp.headers["content-type"] = "text/plain"
    resp.write(body)


def echo(req, resp):
    body = req.read()
    resp.headers["content-length"] = len(body)
    resp.headers["content-type"] = "text/plain"
    resp.write(body)


def sendfile(req, resp):
    path = req.path
    resp.headers["content-length"] = os.path.getsize(path)
    resp.headers["content-type"] = "text/plain"
    with open(path, "rb") as f:
        resp.write(f.read())


def log_response(resp):
    pprint.pprint((resp.status, resp.reason, resp.getheaders()))


@contextmanager
def make_connection(server):
    con = uhttp.UnixHTTPConnection(server.server_address, timeout=2)
    with closing(con):
        yield con



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import json
import sys
import threading
import time

from .. _internal import util


FORMAT_TEXT = "text"
FORMAT_JSON = "json"

DEFAULT_WIDTH = 79


class OutputFormat:

    def __init__(self, output, width=DEFAULT_WIDTH):
        """
        Arguments:
            output (fileobj): file to write progress to (default sys.stdout).
            width (int): width of progress bar in characters (default 79)
        """
        self._output = output
        # TODO: use current terminal width instead.
        self._width = width

    def _write_line(self, line, end="\n"):
        self._output.write(line + end)
        self._output.flush()

    def draw(self, elapsed, value, transferred, size=None,
             phase=None, last=False):
        raise NotImplementedError


class TextFormat(OutputFormat):

    def draw(self, elapsed, value, transferred, size=None,
             phase=None, last=False):
        progress = f"{max(0, value):3d}%" if size else "----"
        done = util.humansize(transferred)
        rate = util.humansize((transferred / elapsed) if elapsed else 0)
        phase = f" | {phase}" if phase else ""
        line = f"[ {progress} ] {done}, {elapsed:.2f} s, {rate}/s{phase}"

        # Using "\r" moves the cursor to the first column, so the next progress
        # will overwrite this one. If this is the last progress, we use "\n" to
        # move to the next line. Otherwise, the next shell prompt will include
        # part of the old progress.
        end = "\n" if last else "\r"

        line = line.ljust(self._width, " ")
        self._write_line(line, end)


class JsonFormat(OutputFormat):

    def draw(self, elapsed, value, transferred, size=None,
             phase=None, last=False):
        progress = {
            'transferred': transferred,
            'elapsed': elapsed,
            'description': phase or "",
        }
        if size is not None:
            progress["size"] = size

        line = json.dumps(progress)
        self._write_line(line)


OUTPUT_FORMAT = {
    FORMAT_TEXT: TextFormat,
    FORMAT_JSON: JsonFormat,
}


class ProgressBar:

    def __init__(self, phase=None, error_phase="command failed", size=None,
                 output=sys.stdout, format=FORMAT_TEXT, step=None,
                 width=DEFAULT_WIDTH, now=time.monotonic):
        """
        Arguments:
            phase (str): short description of the current phase.
            error_phase (str): phase to set when code run under the context
                manager has failed.
            size (int): total number of bytes. If size is unknown when
                creating, progress value is not displayed. The size can be set
                later to enable progress display.
            output (fileobj): file to write progress to (default sys.stdout).
            format (str): format in which the progress is printed.
            step (float): unused, kept for backward compatibility. The progress
                is updated in 1 percent steps.
            width (int): width of progress bar in characters (default 79)
            now (callable): callable returning current time for testing.
        """
        self._phase = phase
        self._error_phase = error_phase
        self._size = size
        self._output = output
        self._format = OUTPUT_FORMAT[format](output=output, width=width)
        self._now = now
        self._lock = threading.Lock()
        self._start = self._now()
        # Number of bytes transferred.
        self._done = 0
        # Value in percent. We start with -1 so the first update will show 0%.
        self._value = -1
        self._closed = False

        # The first update can take some time.
        self._draw()

    @property
    def phase(self):
        return self._phase

    @phase.setter
    def phase(self, s):
        with self._lock:
            if self._closed:
                return
            if self._phase != s:
                self._phase = s
                self._draw()

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, n):
        with self._lock:
            if self._closed:
                return
            if self._size != n:
                self._size = n
                self._draw()

    def update(self, n):
        """
        Increment the progress by n bytes.

        Should be called for every successful transfer, with the number of
        bytes transferred. The number of bytes transferred is updated on every
        call, but the progress is drawn only when the value changes.

        Note: this interface is compatible with tqdm[1], to allow user to use
        different progress implementations.

        [1] https://github.com/tqdm/tqdm#manual
        """
        with self._lock:
            if self._closed:
                return
            self._done += n
            if self._size:
                new_value = int(self._done / self._size * 100)
                if new_value > self._value:
                    self._value = new_value
                    self._draw()

    def close(self):
        with self._lock:
            if not self._closed:
                self._closed = True
                self._draw(last=True)

    def _draw(self, last=False):
        self._format.draw(
            elapsed=self._now() - self._start,
            value=self._value,
            transferred=self._done,
            size=self._size,
            phase=self._phase,
            last=last
        )

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        # If an exception was raised in the caller code, show a failure.
        if t is not None:
            self.phase = self._error_phase
        self.close()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import bisect


class Range:

    __slots__ = ("start", "end")

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __lt__(self, other):
        if self.start < other.start:
            return True
        if self.start == other.start:
            return self.end < other.end
        return False

    def __len__(self):
        return self.end - self.start

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and
                self.start == other.start and
                self.end == other.end)

    def __repr__(self):
        return f"Range(start={self.start}, end={self.end})"


class RangeList:

    def __init__(self, other=None):
        if other:
            self._ranges = [Range(r.start, r.end) for r in other._ranges]
        else:
            self._ranges = []

    def add(self, r):
        """
        Add a single range.
        """
        bisect.insort_left(self._ranges, r)
        self._ranges = _merged(self._ranges)

    def update(self, rs):
        """
        Update from iterable of unsorted ranges.
        """
        self._ranges.extend(rs)
        self._ranges.sort()
        self._ranges = _merged(self._ranges)

    def sum(self):
        return sum(len(r) for r in self._ranges)


def _merged(ranges):
    """
    Merge sorted list of ranges.

    The ranges are sorted, but may contain:

    - consecutive ranges (very likely):

      [Range(0, 100), Range(100, 200)]

    - duplicate ranges (unlikely):

      [Range(0, 100), Range(0, 100)]

    - ranges overlapping other ranges (unlikely):

      [Range(0, 100), Range(50, 200)]

    Return a merged list of ranges without duplicates, overlaps, or
    consecutive ranges.
    """
    merged = ranges[:1]

    for r in ranges[1:]:
        current = merged[-1]
        if current.end >= r.start:
            current.end = max(current.end, r.end)
        else:
            merged.append(r)

    return merged



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
io - I/O operations on backends.
"""

import logging
import threading

from collections import deque, namedtuple
from contextlib import closing
from functools import partial

from .. _internal import util
from .. _internal.backends import Wrapper
from .. _internal.units import MiB

from . import _app

# Limit maximum zero and copy size to spread the workload better to multiple
# workers and ensure frequent progress updates when handling large extents.
MAX_ZERO_SIZE = 128 * MiB
MAX_COPY_SIZE = 128 * MiB

# NBD hard limit.
MAX_BUFFER_SIZE = 32 * MiB

# TODO: Needs more testing.
BUFFER_SIZE = 4 * MiB
MAX_WORKERS = 4

log = logging.getLogger("io")


def copy(src, dst, dirty=False, max_workers=MAX_WORKERS,
         buffer_size=BUFFER_SIZE, zero=True, hole=True, progress=None,
         name="copy"):

    buffer_size = min(buffer_size, MAX_BUFFER_SIZE)

    with Executor(name=name) as executor:
        # This is a bit ugly. We get src and dst backends, to keep same
        # interface as the non-concurrent version. We use src backend here to
        # iterate over image extents. We need to clone src backend max_workers
        # times, and dst backend max_workers - 1) times.

        # The first worker clones src and use a wrapped dst.
        executor.add_worker(
            partial(Handler, src.clone, lambda: Wrapper(dst), buffer_size,
                    progress))

        # The rest of the workers clone both src and dst.
        for _ in range(max_workers - 1):
            executor.add_worker(
                partial(Handler, src.clone, dst.clone, buffer_size, progress))

        if progress:
            progress.size = src.size()

        try:
            # Submit requests to executor.
            if dirty:
                _copy_dirty(executor, src, progress=progress)
            else:
                _copy_data(
                    executor, src, zero=zero, hole=hole, progress=progress)
        except Closed:
            # Error will be raised when exiting the context.
            log.debug("Executor failed")


def _copy_dirty(executor, src, progress=None):
    """
    Copy dirty extents, skipping clean extents. Since we always write to new
    empty qcow2 image, clean areas are unallocated, exposing data from backing
    chain.
    """
    for ext in src.extents("dirty"):
        if ext.dirty:
            if ext.data:
                log.debug("Copying %s", ext)
                executor.submit(Request(COPY, ext.start, ext.length))
            elif ext.zero:
                log.debug("Zeroing %s", ext)
                executor.submit(Request(ZERO, ext.start, ext.length))
        else:
            log.debug("Skipping %s", ext)
            if progress:
                progress.update(ext.length)


def _copy_data(executor, src, zero=True, hole=True, progress=None):
    """
    Copy data extents and zero zero and hole extents.

    The defaults are correct when copying to raw or qcow2 image without a
    backing file, when we do not know if the destination image is empty. If the
    destination image is raw, the backend is sparse, and the storage supports
    punching holes, zeroing will deallocate space. With qcow2 format, areas in
    the qcow2 are never deallocated when zeroing.

    When copying to qcow2 image with a backing file, holes must not be zeroed,
    since zeroed areas will hide data from the backing chain. Use hole=False to
    skip holes and keep them unallocated in the destination image.

    When copying to new empty image without a backing file, we can optimize the
    copy. Use zero=False to skip both zero and hole extents and leave the area
    unallocated.
    """
    for ext in src.extents("zero"):
        if ext.data:
            log.debug("Copying %s", ext)
            executor.submit(Request(COPY, ext.start, ext.length))
        elif zero and (not ext.hole or hole):
            log.debug("Zeroing %s", ext)
            executor.submit(Request(ZERO, ext.start, ext.length))
        else:
            log.debug("Skipping %s", ext)
            if progress:
                progress.update(ext.length)


# Request ops.
ZERO = "zero"
COPY = "copy"
STOP = "stop"


class Request(namedtuple("Request", "op,start,length")):

    def __new__(cls, op, start=0, length=0):
        return tuple.__new__(cls, (op, start, length))


class Executor:

    def __init__(self, name="executor", queue_depth=32):
        self._name = name
        self._workers = []
        self._queue = Queue(queue_depth)
        self._errors = []

    # Public interface.

    def add_worker(self, handler_factory):
        name = "{}/{}".format(self._name, len(self._workers))
        w = Worker(handler_factory, self._queue, self._errors, name=name)
        self._workers.append(w)

    def submit(self, req):
        """
        Submit request to queue. Blocks if the queue is full.
        """
        for req in self._split(req):
            self._queue.put(req)

    def stop(self):
        """
        Stop the executor when pending requests are processed. Blocks until all
        workers exit, and report the first executor error.
        """
        log.debug("Stopping executor %s", self._name)
        for _ in self._workers:
            try:
                self._queue.put(Request(STOP))
            except Closed:
                break
        self._join_workers()
        if self._errors:
            raise self._errors[0]

    def abort(self):
        """
        Drops pending requests and terminate all workers. Blocks until all
        workers exit.
        """
        log.debug("Aborting executor %s", self._name)
        self._queue.close()
        self._join_workers()

    # Private.

    def _join_workers(self):
        for w in self._workers:
            w.join()

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        if t is None:
            # Normal shutdown.
            self.stop()
        else:
            # Do not hide exception in user context.
            try:
                self.abort()
            except Exception:
                log.exception("Error aborting executor")

    def _split(self, req):
        """
        Spread workload on all workers by splitting large requests.
        """
        step = MAX_ZERO_SIZE if req.op == ZERO else MAX_COPY_SIZE
        start = req.start
        length = req.length

        while length > step:
            yield Request(req.op, start, step)
            start += step
            length -= step

        yield Request(req.op, start, length)


class Worker:

    def __init__(self, handler_factory, queue, errors, name="worker"):
        self._handler_factory = handler_factory
        self._queue = queue
        self._errors = errors
        self._name = name

        log.debug("Starting worker %s", name)
        self._thread = util.start_thread(self._run, name=name)

    def join(self):
        log.debug("Waiting for worker %s", self._name)
        self._thread.join()

    def _run(self):
        try:
            log.debug("Worker %s started", self._name)
            handler = self._handler_factory()
            with closing(handler):
                while True:
                    req = self._queue.get()
                    _app.check_terminated()
                    if req.op is ZERO:
                        handler.zero(req)
                    elif req.op is COPY:
                        handler.copy(req)
                    elif req.op is STOP:
                        handler.flush(req)
                        break
        except Closed:
            log.debug("Worker %s cancelled", self._name)
        except Exception as e:
            self._errors.append(e)
            self._queue.close()
            log.debug("Worker %s failed: %s", self._name, e)
        else:
            log.debug("Worker %s finished", self._name)


class Handler:

    def __init__(self, src_factory, dst_factory, buffer_size=BUFFER_SIZE,
                 progress=None):
        # Connecting to backend server may fail. Don't leave open connections
        # after failures.
        self._src = src_factory()
        try:
            self._dst = dst_factory()
        except Exception:
            self._src.close()
            raise

        self._buf = bytearray(buffer_size)
        self._progress = progress

    def zero(self, req):
        # TODO: Assumes complete zero(); not compatible with file backend.
        self._dst.seek(req.start)
        self._dst.zero(req.length)
        if self._progress:
            self._progress.update(req.length)

    def copy(self, req):
        self._src.seek(req.start)
        self._dst.seek(req.start)

        if hasattr(self._dst, "read_from"):
            self._dst.read_from(self._src, req.length, self._buf)
        elif hasattr(self._src, "write_to"):
            self._src.write_to(self._dst, req.length, self._buf)
        else:
            self._generic_copy(req)

        if self._progress:
            self._progress.update(req.length)

    def flush(self, req):
        self._dst.flush()

    def close(self):
        # Error while closing the destination backend should fail the
        # operation. Error in closing source is not fatal, but we want to know
        # about it.
        try:
            self._dst.close()
        finally:
            try:
                self._src.close()
            except Exception:
                log.exception("Error closing %s", self._src)

    def _generic_copy(self, req):
        # TODO: Assumes complete readinto() and write(); not compatible with
        # file backend.
        step = len(self._buf)
        todo = req.length

        while todo > step:
            self._src.readinto(self._buf)
            self._dst.write(self._buf)
            todo -= step

        with memoryview(self._buf)[:todo] as view:
            self._src.readinto(view)
            self._dst.write(view)


class Closed(Exception):
    """
    Raised when trying to access a closed queue.
    """


class Queue:
    """
    A simple queue supporting cancellation.

    Once a queue is closed, putting items or getting items will raise a Closed
    exception. This makes it easy to cancel group of threads waiting on the
    queue.
    """

    def __init__(self, max_size):
        self._cond = threading.Condition(threading.Lock())
        self._queue = deque(maxlen=max_size)
        self._closed = False

    @property
    def closed(self):
        return self._closed

    def put(self, item):
        with self._cond:
            self._wait_while(length=self._queue.maxlen)
            self._queue.append(item)
            self._cond.notify()

    def get(self):
        with self._cond:
            self._wait_while(length=0)
            item = self._queue.popleft()
            if len(self._queue) == self._queue.maxlen - 1:
                self._cond.notify()
            return item

    def _wait_while(self, length):
        if self._closed:
            raise Closed
        while len(self._queue) == length:
            self._cond.wait()
            if self._closed:
                raise Closed

    def close(self):
        with self._cond:
            self._closed = True
            self._queue.clear()
            self._cond.notify_all()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Smart client for accessing imageio server.

This module exposes the public names. Anything else in this package is private
and should not be used.
"""

# flake8: noqa

from .. _internal import version

# The public APIs
from . _api import (
    BUFFER_SIZE,
    MAX_WORKERS,
    upload,
    download,
    info,
    measure,
    checksum,
    extents,
    ImageioClient,
)

# For better user experience.
from . _ui import ProgressBar

__all__ = (
    "BUFFER_SIZE",
    "MAX_WORKERS",
    "ImageioClient",
    "ProgressBar",
    "checksum",
    "download",
    "extents",
    "info",
    "measure",
    "upload",
)

__version__ = version.string



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
ipv6 - IP v6 helper module
"""


def unquote_address(host):
    """
    Unquote IPv6 quoted numeric address.
    """
    if host.startswith("[") and host.endswith("]") and ":" in host:
        host = host[1:-1]
    return host


def quote_address(host):
    """
    Quote IPv6 numeric address.
    """
    if ":" in host:
        host = "[{}]".format(host)
    return host



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Tool options.
"""

import getpass
from collections import namedtuple

from .. _internal import version
from .. _internal.units import KiB, MiB, GiB, TiB
from . _api import MAX_WORKERS, BUFFER_SIZE

import argparse
import configparser
import os
import sys
import uuid


ADD_DISK_TIMEOUT = 120
# Possible boolean values in the configuration.
BOOLEAN_VALUES = {'1': True, 'yes': True, 'true': True, 'on': True,
                  '0': False, 'no': False, 'false': False, 'off': False}


class Choices:

    def __init__(self, name, values):
        self.name = name
        self.values = values

    def __call__(self, s):
        if s not in self.values:
            raise ValueError(
                f"Invalid '{self.name}' value: '{s}', choices: {self}")
        return s

    def __str__(self):
        s = ", ".join(self.values)
        return f"{{{s}}}"

    def __repr__(self):
        return repr(self.name)


log_level = Choices("log_level", ("debug", "info", "warning", "error"))
output_format = Choices("output_format", ("text", "json"))


def bool_string(value):
    if value.lower() not in BOOLEAN_VALUES:
        raise ValueError(f"Not a boolean: '{value}'")
    return BOOLEAN_VALUES[value.lower()]


class Option(
        namedtuple(
            "Option", "name,args,config,required,action,type,default,help")):

    __slots__ = ()

    def __new__(cls, name=None, args=(), config=False, required=False,
                action=None, type=str, default=None, help=None):
        """
        Arguments:
            name (str): Option name use in config file or parsed arguments.
            args (List[str]): If set, add command line arguments for this
                option.
            config (bool): If True this option can be set in the config file.
            required (bool): If True this option is required and parsing
                arguments will fail if it is not specified in the command line
                or config file.
            action (Union[argparse.Action, str]): Specify how an argument
                should be handled. Defaults to 'store' action.
            type (callable): Callable converting the command line argument or
                config value to the wanted type. Must raise ValueError if the
                value is invalid and cannot be converted.
            default (Any): The default value if the option was not set in the
                command line arguments or config file.
            help (str): Help message to show in the online help.
        """
        return tuple.__new__(
            cls, (name, args, config, required, action, type, default, help))

    @property
    def kwargs(self):
        option_kwargs = {
            "dest": self.name,
            "default": None,
            "help": self.help,
        }
        if self.action:
            option_kwargs["action"] = self.action
        else:
            option_kwargs["type"] = self.type
        return option_kwargs


class Parser:

    _OPTIONS = [
        Option(
            name="config",
            args=["-c", "--config"],
            help="If set, read specified section from the configuration file.",
        ),
        Option(
            name="engine_url",
            args=["--engine-url"],
            config=True,
            required=True,
            help=("ovirt-engine URL. If not set, read from the specified "
                  "config section (required)."),
        ),
        Option(
            name="username",
            args=["--username"],
            config=True,
            required=True,
            help=("ovirt-engine username. If not set, read from the "
                  "specified config section (required)."),
        ),
        Option(
            name="password",
            config=True,
        ),
        Option(
            name="password_file",
            args=["--password-file"],
            help=("Read ovirt-engine password from file. If not set, read "
                  "password from the specified config section, or prompt "
                  "the user for the password."),
        ),
        Option(
            name="cafile",
            args=["--cafile"],
            config=True,
            help=("Path to ovirt-engine CA certificate. If not set, read "
                  "from the specified config section"),
        ),
        Option(
            name="secure",
            args=["--insecure"],
            action="store_false",
            config=True,
            default=True,
            type=bool_string,
            help=("Do not verify server certificates and host name (not "
                  "recommened).")
        ),
        Option(
            name="output",
            args=["-o", "--output"],
            type=output_format,
            default="text",
            help="Set the format in which the output must be done "
                 f"(choices: {output_format}), default: text).",
        ),
        Option(
            name="disk_timeout",
            config=True,
            type=int,
            default=ADD_DISK_TIMEOUT,
        ),
        Option(
            name="log_file",
            args=["--log-file"],
            config=True,
            help="Log to file instead of stderr.",
        ),
        Option(
            name="log_level",
            args=["--log-level"],
            config=True,
            type=log_level,
            default="warning",
            help=(f"Log level (choices: {log_level}, default: warning)."),
        ),
    ]

    def __init__(self):
        self._parser = argparse.ArgumentParser(
            description="Transfer disk images")
        self._parser.set_defaults(command=None)
        self._parser.add_argument(
            '--version',
            action='version',
            version=f'%(prog)s {version.string}')
        self._commands = self._parser.add_subparsers(title="commands")

    def add_sub_command(self, name, help, func, transfer_options=True):
        cmd = self._commands.add_parser(name, help=help)
        cmd.set_defaults(command=func)

        for option in self._OPTIONS:
            # Skip config only options.
            if not option.args:
                continue

            cmd.add_argument(*option.args, **option.kwargs)

        if transfer_options:
            size = Size(minimum=1, default=MAX_WORKERS, maximum=8)
            cmd.add_argument(
                "--max-workers",
                type=size,
                default=size.default,
                help=f"Maximum number of workers (range: {size.minimum}-"
                f"{size.maximum}, default: {size.default}).")

            size = Size(
                minimum=64 * KiB, default=BUFFER_SIZE, maximum=16 * MiB)
            cmd.add_argument(
                "--buffer-size",
                type=size,
                default=size.default,
                help=f"Buffer size per worker (range: {size.minimum}-"
                f"{size.maximum}, default: {size.default}).")

        return cmd

    def parse(self, args=None):
        args = self._parser.parse_args(args=args)
        if not args.command:
            self._parser.print_help()
            sys.exit(2)

        if args.config:
            self._merge_config(args)

        self._set_defaults(args)
        self._check_required_options(args)
        self._read_password(args)

        return args

    @property
    def config_file(self):
        # https://specifications.freedesktop.org/basedir-spec
        base_dir = os.environ.get("XDG_CONFIG_HOME")
        if not base_dir:
            base_dir = os.path.expanduser("~/.config")
        return os.path.join(base_dir, "ovirt-img.conf")

    def _merge_config(self, args):
        """
        Read options from specified config section and set the value for unused
        command line options. This must be done after parsing command line
        options since we depend on --config.
        """
        config = configparser.ConfigParser(interpolation=None)
        config.read([self.config_file])

        if not config.has_section(args.config):
            self._parser.error(
                f"No section: '{args.config}' in '{self.config_file}'")

        for option in self._OPTIONS:
            # Skip command line only options.
            if not option.config:
                continue

            # Skip options set by the command line.
            if getattr(args, option.name, None) is not None:
                continue

            # Read and validate value from config file.
            if config.has_option(args.config, option.name):
                raw_value = config.get(args.config, option.name)
                try:
                    value = option.type(raw_value)
                except ValueError as e:
                    self._parser.error(e)
                setattr(args, option.name, value)

    def _set_defaults(self, args):
        """
        Add default value for unused command line argument (None by default),
        and unset config only values (attribute missing). This must be done
        after reading config values.
        """
        for option in self._OPTIONS:
            if getattr(args, option.name, None) is None:
                setattr(args, option.name, option.default)

    def _check_required_options(self, args):
        """
        Check that all required options have some value. This must be done
        after setting the defaults.
        """
        missing = [opt.name for opt in self._OPTIONS
                   if opt.required and getattr(args, opt.name) is None]
        if missing:
            self._parser.error(
                f"Missing required options: {', '.join(missing)}")

    def _read_password(self, args):
        """
        Try to read the password from the args.password_file, args.password, or
        prompt the user for the password.
        """
        if args.password_file:
            with open(args.password_file) as f:
                args.password = f.read().rstrip("\n")
        elif args.password is None:
            args.password = getpass.getpass()


class SizeValue(int):

    SUFFIXES = {"": 1, "k": KiB, "m": MiB, "g": GiB, "t": TiB}

    def __str__(self):
        n = int(self)
        for unit in self.SUFFIXES:
            if n < KiB:
                break
            n //= KiB
        return f"{n}{unit}"


class Size:
    """
    Convert and validate size string.
    """

    def __init__(self, minimum=0, default=None, maximum=None):
        # Minimum value is required since negative size does not make sense.
        self.minimum = SizeValue(minimum)
        self.default = None if default is None else SizeValue(default)
        self.maximum = None if maximum is None else SizeValue(maximum)

    def __call__(self, s):
        if s == "":
            raise ValueError(f"Invalid size: {s!r}")

        unit = SizeValue.SUFFIXES.get(s[-1].lower())
        if unit:
            value = int(s[:-1]) * unit
        else:
            value = int(s)

        if value < self.minimum:
            raise ValueError(f"Size {s!r} < {self.minimum}")

        if self.maximum is not None and value > self.maximum:
            raise ValueError(f"Size {s!r} > {self.maximum}")

        return value


class Type:
    """
    Wrapper to expose function validators as types.
    """

    def __init__(self, name, func):
        self.name = name
        self.func = func

    def __call__(self, value):
        try:
            return self.func(value)
        except ValueError as e:
            raise ValueError(
                f"Invalid '{self.name}' value: '{value}': {e}") from e

    def __repr__(self):
        return self.name


def _validate_uuid(id):
    return str(uuid.UUID(id))


def _validate_file(filename):
    """
    Validate that the file exists.

    Raises:
        ValueError

    Returns:
        str: filename
    """
    if not os.path.exists(filename):
        raise ValueError(f"{filename} does not exist")
    if not os.path.isfile(filename):
        raise ValueError(f"{filename} is not a file")
    return filename


UUID = Type("UUID", _validate_uuid)
File = Type("File", _validate_file)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

class Closed:

    def __getattr__(self, name):
        """
        Behave like closed io.FileIO.
        """
        raise ValueError("Operation on closed backend")


CLOSED = Closed()



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

from .. import cors
from .. import version


class Handler:
    """
    Handle requests for the /info resource.
    """

    def __init__(self, config, auth):
        self.config = config
        self.auth = auth

    @cors.allow()
    def get(self, req, resp):
        resp.send_json({"version": version.string})



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import pytest

from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd


@pytest.mark.parametrize("src_fmt,dst_fmt", [
    ("raw", "raw"),
    ("qcow2", "qcow2"),
    ("raw", "qcow2"),
])
def test_compare_identical_content(tmpdir, src_fmt, dst_fmt):
    size = 1024**2
    src = str(tmpdir.join("src." + src_fmt))
    dst = str(tmpdir.join("dst." + dst_fmt))

    qemu_img.create(src, src_fmt, size=size)
    qemu_img.create(dst, dst_fmt, size=size)

    # Destination image has different allocation.
    with qemu_nbd.open(dst, dst_fmt) as c:
        c.write(size // 2, b"\0")
        c.flush()

    qemu_img.compare(src, dst, format1=src_fmt, format2=dst_fmt)


@pytest.mark.parametrize("src_fmt,dst_fmt", [
    ("raw", "raw"),
    ("qcow2", "qcow2"),
    ("raw", "qcow2"),
])
def test_compare_different_content(tmpdir, src_fmt, dst_fmt):
    size = 1024**2
    src = str(tmpdir.join("src." + src_fmt))
    dst = str(tmpdir.join("dst." + dst_fmt))

    qemu_img.create(src, src_fmt, size=size)
    qemu_img.create(dst, dst_fmt, size=size)

    # Destination image has different content.
    with qemu_nbd.open(dst, dst_fmt) as c:
        c.write(size // 2, b"x")
        c.flush()

    with pytest.raises(qemu_img.ContentMismatch):
        qemu_img.compare(src, dst, format1=src_fmt, format2=dst_fmt)


def test_compare_wrong_format(tmpdir):
    size = 1024**2
    src = str(tmpdir.join("src.raw"))
    dst = str(tmpdir.join("dst.raw"))

    qemu_img.create(src, "raw", size=size)
    qemu_img.create(dst, "raw", size=size)

    with pytest.raises(qemu_img.OpenImageError):
        qemu_img.compare(src, dst, format1="qcow2")

    with pytest.raises(qemu_img.OpenImageError):
        qemu_img.compare(src, dst, format2="qcow2")


@pytest.mark.parametrize("src_fmt,dst_fmt", [
    ("raw", "raw"),
    ("qcow2", "qcow2"),
    ("raw", "qcow2"),
])
def test_compare_different_allocation(tmpdir, src_fmt, dst_fmt):
    # Images has same content, but different allocation.
    size = 1024**2
    src = str(tmpdir.join("src." + src_fmt))
    dst = str(tmpdir.join("dst." + dst_fmt))

    qemu_img.create(src, src_fmt, size=size)
    qemu_img.create(dst, dst_fmt, size=size)

    with qemu_nbd.open(dst, dst_fmt) as c:
        c.write(size // 2, b"\0")
        c.flush()

    with pytest.raises(qemu_img.ContentMismatch):
        qemu_img.compare(
            src, dst, format1=src_fmt, format2=dst_fmt, strict=True)


def test_compare_missing_file(tmpdir):
    src = str(tmpdir.join("src.raw"))
    dst = str(tmpdir.join("dst.raw"))

    qemu_img.create(src, "raw", size=1024**2)

    with pytest.raises(qemu_img.OpenImageError):
        qemu_img.compare(src, dst)


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_create_info(tmpdir, fmt):
    size = 1024**2
    image = str(tmpdir.join("image." + fmt))
    qemu_img.create(image, fmt, size=size)
    info = qemu_img.info(image)

    assert info["filename"] == image
    assert info["virtual-size"] == size
    assert info["format"] == fmt


def test_add_bitmap(tmpdir):
    size = 10 * 1024**2
    img = str(tmpdir.join("img.qcow2"))
    qemu_img.create(img, "qcow2", size=size)
    qemu_img.bitmap_add(img, "b0")
    bitmaps = qemu_img.info(img)["format-specific"]["data"]["bitmaps"]
    assert bitmaps == [
        {"name": "b0", "flags": ["auto"], "granularity": 65536}
    ]



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import http.client as http_client
import logging
import os
import socket
import uuid

from . import http
from . import util

PUT = "PUT"
DELETE = "DELETE"
PATCH = "PATCH"
GET = "GET"

log = logging.getLogger("uhttp")


class UnsupportedError(Exception):
    pass


class _UnixMixin:

    def set_tunnel(self, host, port=None, headers=None):
        raise UnsupportedError("Tunneling is not supported")


class UnixHTTPConnection(_UnixMixin, http_client.HTTPConnection):
    """
    HTTP connection over unix domain socket.
    """

    def __init__(self, path, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        self.path = path
        super().__init__("localhost", timeout=timeout)

    def connect(self):
        self.sock = _create_unix_socket(self.timeout)
        self.sock.connect(self.path)


class Server(http.Server):
    """
    HTTP server over unix domain socket.
    """

    address_family = socket.AF_UNIX
    server_name = "localhost"
    server_port = None

    def create_socket(self, prefer_ipv4=False):
        self.socket = socket.socket(socket.AF_UNIX, self.socket_type)

        if self.server_address == "":
            # User wants to bind to a random abstract socket.
            # A recent change in cpython broke autobind of abstract
            # unix socket in linux, causing bind calls with an
            # empty string to be bound to '\0' instead of a
            # random address.
            # Until it is fixed, we need to generate random address ourselves.
            # See https://github.com/python/cpython/issues/94821
            self.server_address = f"\0{uuid.uuid4()}"

    def server_bind(self):
        """
        Override to remove existing socket for pathname sockets, and support
        random abstract sockets.  See unix(7) for details.
        """
        if self.server_address and self.server_address[0] != "\0":
            # A pathname socket must be removed before binding.
            self._remove_socket()

        self.socket.bind(self.server_address)
        # Socket addresses in Linux abstract namespace are returned as bytes,
        # so we have to eventually convert address to string.
        # See https://docs.python.org/3.9/library/socket.html#socket-families
        self.server_address = util.ensure_text(self.socket.getsockname())

    def get_request(self):
        """
        Override to return non-empty client address, expected by
        WSGIRequestHandler.
        """
        sock, _ = self.socket.accept()
        return sock, self.server_address

    def shutdown(self):
        if self.server_address[0] != "\0":
            self._remove_socket()
        super().shutdown()

    def _remove_socket(self):
        log.debug("Removing socket %r", self.server_address)
        try:
            os.unlink(self.server_address)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise


class Connection(http.Connection):
    """
    HTTP connection over unix domain socket.
    """

    # Not needed for unix socket.
    disable_nagle_algorithm = False

    def address_string(self):
        """
        Override to return meaningfull string instead of the first character of
        the unix socket path.
        """
        return "local"


def _create_unix_socket(timeout):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    if timeout is not socket._GLOBAL_DEFAULT_TIMEOUT:
        sock.settimeout(timeout)
    return sock



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import hashlib
import json
import os

import userstorage
import pytest

from ovirt_imageio._internal import blkhash
from ovirt_imageio._internal import config
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd
from ovirt_imageio._internal import server

from .. import testutil
from .. import http

BACKENDS = userstorage.load_config("storage.py").BACKENDS
ALGORITHMS = frozenset(hashlib.algorithms_available)


@pytest.fixture(
    params=[
        BACKENDS["file-512-xfs"],
        BACKENDS["file-4k-xfs"],
    ],
    ids=str
)
def user_file(request):
    backend = request.param
    with backend:
        yield backend


@pytest.fixture(scope="module")
def srv():
    cfg = config.load(["test/conf/daemon.conf"])
    s = server.Server(cfg)
    s.start()
    yield s
    s.stop()


@pytest.fixture(params=[
    pytest.param(http.RemoteClient, id="http"),
    pytest.param(http.LocalClient, id="local"),
])
def client(srv, request):
    srv.auth.clear()
    client = request.param(srv.config)
    yield client
    client.close()


def test_unsupported_method(srv, client):
    res = client.request("PUT", "/images/ticket-id/checksum")
    assert res.status == 405


def test_no_ticket_id(srv, client):
    res = client.request("GET", "/images//checksum")
    assert res.status == 400


def test_no_ticket(srv, client):
    res = client.request("GET", "/images/no-such-ticket/checksum")
    assert res.status == 403


def test_ticket_expired(srv, client, fake_time):
    ticket = testutil.create_ticket(timeout=300)
    srv.auth.add(ticket)

    # Make the ticket expire
    fake_time.now += 300

    res = client.request("GET", "/images/%(uuid)s/checksum" % ticket)
    assert res.status == 403


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_file_backend(srv, client, user_file, fmt):
    qemu_img.create(user_file.path, fmt, size="2m")

    with qemu_nbd.open(user_file.path, fmt) as c:
        # Add cluster with data.
        c.write(1 * 1024**2, b"some data")
        c.flush()

    # File backend operate on host data, not guest data.
    checksum = blkhash.checksum(user_file.path)

    # File backend uses actual size, not vitual size.
    size = os.path.getsize(user_file.path)

    ticket = testutil.create_ticket(url="file://" + user_file.path, size=size)
    srv.auth.add(ticket)

    res = client.request("GET", "/images/{}/checksum".format(ticket["uuid"]))
    data = res.read()
    assert res.status == 200

    res = json.loads(data)
    assert res == checksum


@pytest.mark.parametrize("block_size", [
    blkhash.BLOCK_SIZE // 4,
    blkhash.BLOCK_SIZE,
    blkhash.BLOCK_SIZE + 4096,
    blkhash.BLOCK_SIZE * 4,
])
def test_user_block_size(srv, client, tmpdir, block_size):
    size = 2 * 1024**2

    img = str(tmpdir.join("file"))
    qemu_img.create(img, "raw", size=size)

    # File backend operate on host data, not guest data.
    checksum = blkhash.checksum(img, block_size=block_size)

    ticket = testutil.create_ticket(url="file://" + img, size=size)
    srv.auth.add(ticket)

    res = client.request("GET", "/images/{}/checksum?block_size={}"
                         .format(ticket["uuid"], block_size))
    data = res.read()
    assert res.status == 200

    res = json.loads(data)
    assert res == checksum


@pytest.mark.parametrize("block_size", [
    "invalid",
    blkhash.BLOCK_SIZE // 4 - 1,
    blkhash.BLOCK_SIZE * 4 + 1,
    blkhash.BLOCK_SIZE + 1,
])
def test_user_block_size_invalid(srv, client, block_size):
    ticket = testutil.create_ticket(url="file:///no/such/file")
    srv.auth.add(ticket)
    res = client.request("GET", "/images/{}/checksum?block_size={}"
                         .format(ticket["uuid"], block_size))
    res.read()
    assert res.status == 400


@pytest.mark.parametrize("fmt,compressed", [
    ("raw", False),
    ("qcow2", False),
    ("qcow2", True),
])
def test_nbd_backend(srv, client, tmpdir, nbd_server, fmt, compressed):
    size = 2 * 1024**2

    # Create temporary file with some data.
    tmp = tmpdir.join("tmp")
    with open(tmp, "wb") as f:
        f.truncate(size)
        # Add zero allocated cluster.
        f.write(b"\0" * 64 * 1024)
        # Add cluster with data.
        f.seek(1 * 1024**2)
        f.write(b"some data")

    # NBD backend checksums guest visible data.
    checksum = blkhash.checksum(tmp)

    # Create test image.
    qemu_img.convert(tmp, nbd_server.image, "raw", fmt, compressed=compressed)

    nbd_server.fmt = fmt
    nbd_server.start()

    ticket = testutil.create_ticket(url=nbd_server.sock.url(), size=size)
    srv.auth.add(ticket)

    res = client.request("GET", "/images/{}/checksum".format(ticket["uuid"]))
    data = res.read()
    assert res.status == 200

    res = json.loads(data)
    assert res == checksum


@pytest.mark.parametrize("algorithm,digest_size", [
    ("blake2b", 32),
    ("sha1", None),
])
def test_algorithms(srv, client, tmpdir, algorithm, digest_size):
    size = 2 * 1024**2

    image = str(tmpdir.join("image"))
    qemu_img.create(image, "raw", size=size)
    ticket = testutil.create_ticket(url="file://" + image, size=size)
    srv.auth.add(ticket)

    # File backend operate on host data, not guest data.
    checksum = blkhash.checksum(
        image, algorithm=algorithm, digest_size=digest_size)

    path = "/images/{}/checksum?algorithm={}".format(
        ticket["uuid"], algorithm)

    res = client.request("GET", path)
    data = res.read()
    assert res.status == 200

    res = json.loads(data)
    assert res == checksum


def test_algorithms_unknown(srv, client, tmpdir):
    ticket = testutil.create_ticket(url="file:///no/such/file", size=4096)
    srv.auth.add(ticket)
    res = client.request(
        "GET",
        "/images/{}/checksum?algorithm=unknown".format(ticket["uuid"]))
    error = res.read().decode("utf-8")
    assert res.status == 400
    assert repr("unknown") in error
    assert repr("algorithm") in error
    for name in ALGORITHMS:
        assert repr(name) in error


def test_checksum_algorithms(srv, client, tmpdir):
    res = client.request("GET", "/images/no-ticket/checksum/algorithms")
    data = res.read()
    assert res.status == 200

    res = json.loads(data)
    assert res == {"algorithms": sorted(ALGORITHMS)}



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import io
import logging
import os

import pytest
import userstorage

from urllib.parse import urlparse

from ovirt_imageio._internal import nbd
from ovirt_imageio._internal import nbdutil
from ovirt_imageio._internal import qemu_img
from ovirt_imageio._internal import qemu_nbd

from . import backup
from . import ci
from . import distro
from . import testutil

from . marks import (
    flaky_in_ovirt_ci,
    requires_ipv6,
)

BACKENDS = userstorage.load_config("storage.py").BACKENDS

log = logging.getLogger("test")


@pytest.fixture(
    params=[
        BACKENDS["file-512-ext4"],
        BACKENDS["file-512-xfs"],
        BACKENDS["file-4k-ext4"],
        BACKENDS["file-4k-xfs"],
    ],
    ids=str
)
def user_file(request):
    backend = request.param
    with backend:
        yield backend


@pytest.fixture(params=[
    "unix",
    pytest.param("tcp", marks=flaky_in_ovirt_ci),
])
def nbd_sock(request, tmpdir):
    if request.param == "unix":
        return nbd.UnixAddress(tmpdir.join("sock"))
    else:
        return nbd.TCPAddress("localhost", testutil.random_tcp_port())


# Addresses

@pytest.mark.parametrize("addr,export,url", [
    # Note: We get Unicode when parsing ticket JSON.
    (nbd.UnixAddress("/sock"), None, "nbd:unix:/sock"),
    (nbd.UnixAddress("/sock"), "", "nbd:unix:/sock"),
    (nbd.UnixAddress("/sock"), "sda", "nbd:unix:/sock:exportname=sda"),
    (nbd.TCPAddress("host", 10900), None, "nbd:host:10900"),
    (nbd.TCPAddress("host", 10900), "", "nbd:host:10900"),
    (nbd.TCPAddress("host", 10900), "sdb", "nbd:host:10900:exportname=sdb"),
])
def test_url(addr, export, url):
    assert addr.url(export) == url


@pytest.mark.parametrize("host,port", [
    ("localhost", "42"),
    (42, 42),
])
def test_tcp_address_invalid(host, port):
    with pytest.raises(ValueError):
        nbd.TCPAddress(host, port)


# Communicate with qemu-nbd


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
@pytest.mark.parametrize("export", [
    "",
    "ascii",
    pytest.param("\u05d0", id="unicode"),
])
def test_handshake(tmpdir, export, fmt):
    image = str(tmpdir.join("image"))
    create_image(image, fmt, 1024**3)
    sock = nbd.UnixAddress(tmpdir.join("sock"))

    with qemu_nbd.run(image, fmt, sock, export_name=export):
        if export:
            c = nbd.Client(sock, export)
        else:
            c = nbd.Client(sock)
        with c:
            # TODO: test transmission_flags?
            assert c.export_size == 1024**3
            assert c.minimum_block_size == 1
            assert c.preferred_block_size == 4096
            assert c.maximum_block_size == 32 * 1024**2
            # Both available in in current qemu-nbd (>= 5.2.0).
            assert c.has_base_allocation
            assert c.has_allocation_depth == (fmt != "raw")


def test_raw_read(tmpdir):
    image = str(tmpdir.join("image"))
    sock = nbd.UnixAddress(tmpdir.join("sock"))
    offset = 1024**2
    data = b"can read from raw"

    with open(image, "wb") as f:
        f.truncate(1024**3)
        f.seek(offset)
        f.write(data)

    with qemu_nbd.run(image, "raw", sock):
        with nbd.Client(sock) as c:
            assert c.read(offset, len(data)) == data


def test_raw_readinto(nbd_server):
    offset = 1024**2
    data = b"can read from raw"

    with io.open(nbd_server.image, "wb") as f:
        f.truncate(2 * 1024**2)
        f.seek(1024**2)
        f.write(data)

    nbd_server.start()

    with nbd.open(nbd_server.url) as c:
        buf = bytearray(len(data))
        c.readinto(offset, buf)
        assert buf == data


def test_raw_write(tmpdir):
    image = str(tmpdir.join("image"))
    with open(image, "wb") as f:
        f.truncate(1024**3)
    sock = nbd.UnixAddress(tmpdir.join("sock"))
    offset = 1024**2
    data = b"can write to raw"

    with qemu_nbd.run(image, "raw", sock):
        with nbd.Client(sock) as c:
            c.write(offset, data)
            c.flush()

    with open(image, "rb") as f:
        f.seek(offset)
        assert f.read(len(data)) == data


def test_qcow2_write_read(tmpdir):
    image = str(tmpdir.join("image"))
    sock = nbd.UnixAddress(tmpdir.join("sock"))
    offset = 1024**2
    data = b"can read and write qcow2"
    create_image(image, "qcow2", 1024**3)

    with qemu_nbd.run(image, "qcow2", sock):
        with nbd.Client(sock) as c:
            c.write(offset, data)
            c.flush()

        with nbd.Client(sock) as c:
            assert c.read(offset, len(data)) == data


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_zero(tmpdir, fmt):
    size = 2 * 1024**2
    offset = 1024**2
    image = str(tmpdir.join("image"))
    sock = nbd.UnixAddress(tmpdir.join("sock"))
    create_image(image, fmt, size)

    with qemu_nbd.run(image, fmt, sock):
        # Fill image with data
        with nbd.Client(sock) as c:
            c.write(0, b"x" * size)
            c.flush()

        # Zero a range
        with nbd.Client(sock) as c:
            c.zero(offset, 4096)
            c.flush()

        with nbd.Client(sock) as c:
            assert c.read(offset, 4096) == b"\0" * 4096


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_zero_max_block_size(tmpdir, fmt):
    offset = 1024**2
    image = str(tmpdir.join("image"))
    sock = nbd.UnixAddress(tmpdir.join("sock"))
    create_image(image, fmt, 1024**3)

    with qemu_nbd.run(image, fmt, sock):
        # Fill range with data
        with nbd.Client(sock) as c:
            size = c.maximum_block_size
            c.write(offset, b"x" * size)
            c.flush()

        # Zero range using maximum block size
        with nbd.Client(sock) as c:
            c.zero(offset, size)
            c.flush()

        with nbd.Client(sock) as c:
            assert c.read(offset, size) == b"\0" * size


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_zero_min_block_size(tmpdir, fmt):
    offset = 1024**2
    image = str(tmpdir.join("image"))
    sock = nbd.UnixAddress(tmpdir.join("sock"))
    create_image(image, fmt, 1024**3)

    with qemu_nbd.run(image, fmt, sock):
        # Fill range with data
        with nbd.Client(sock) as c:
            size = c.minimum_block_size
            c.write(offset, b"x" * size)
            c.flush()

        # Zero range using minimum block size
        with nbd.Client(sock) as c:
            c.zero(offset, size)
            c.flush()

        with nbd.Client(sock) as c:
            assert c.read(offset, size) == b"\0" * size


def test_zero_allocation(user_file):
    # Create sparse image.
    with open(user_file.path, "w") as f:
        f.truncate(2 * 1024**2)

    with qemu_nbd.open(user_file.path, "raw") as c:
        # Zero with punch_hole=False allocates space.
        c.zero(0, c.export_size, punch_hole=False)
        c.flush()
        assert os.stat(user_file.path).st_blocks * 512 == c.export_size

        # Default zero deallocates space (depends on qemu-nbd default).
        c.zero(0, c.export_size)
        c.flush()
        assert os.stat(user_file.path).st_blocks == 0


@pytest.mark.parametrize("url,export", [
    # Note: We get Unicode URL when parsing ticket JSON.
    ("nbd:unix:/path", ""),
    ("nbd:unix:/path:exportname=", ""),
    ("nbd:unix:/path:exportname=sda", "sda"),
    ("nbd:unix:/path:exportname=/sda", "/sda"),
])
def test_open_unix(tmpdir, url, export):
    image = str(tmpdir.join("image"))
    with open(image, "wb") as f:
        f.truncate(1024**3)

    sock = nbd.UnixAddress(tmpdir.join("sock"))
    url = url.replace("/path", sock)

    log.debug("Trying url=%r export=%r", url, export)
    with qemu_nbd.run(image, "raw", sock, export_name=export):
        with nbd.open(urlparse(url)) as c:
            assert c.export_size == 1024**3


@flaky_in_ovirt_ci
@pytest.mark.parametrize("url_template,host,export", [
    # Note: We get Unicode URL when parsing ticket JSON.
    # DNS name
    ("nbd:localhost:{port}", "localhost", ""),
    ("nbd:localhost:{port}:exportname=", "localhost", ""),
    ("nbd:localhost:{port}:exportname=sda", "localhost", "sda"),
    ("nbd:localhost:{port}:exportname=/sda", "localhost", "/sda"),
    ("nbd://localhost:{port}", "localhost", ""),
    ("nbd://localhost:{port}/", "localhost", ""),
    ("nbd://localhost:{port}/sda", "localhost", "sda"),
    ("nbd://localhost:{port}//sda", "localhost", "/sda"),
    # IPv4
    ("nbd://127.0.0.1:{port}", "127.0.0.1", ""),
    ("nbd:127.0.0.1:{port}", "127.0.0.1", ""),
    # IPv6
    pytest.param("nbd://[::1]:{port}", "[::1]", "", marks=requires_ipv6),
    pytest.param("nbd:[::1]:{port}", "[::1]", "", marks=requires_ipv6),
])
def test_open_tcp(tmpdir, url_template, host, export):
    image = str(tmpdir.join("image"))
    with open(image, "wb") as f:
        f.truncate(1024**3)

    sock = nbd.TCPAddress(host, testutil.random_tcp_port())
    url = url_template.format(port=sock.port)

    log.debug("Trying url=%r export=%r", url, export)
    with qemu_nbd.run(image, "raw", sock, export_name=export):
        with nbd.open(urlparse(url)) as c:
            assert c.export_size == 1024**3


FMT_HOLE_FLAGS_PARAMS = [
    # In qemu-nbd 5.2.0 we get only the zero bit.
    ["raw", nbd.STATE_ZERO],
    ["qcow2", nbd.STATE_ZERO | nbd.STATE_HOLE | nbd.EXTENT_BACKING],
]

# Since qemu-nbd 6.0.0 we get also the hole bit, but we don't get the backing
# bit, so unallocated areas in raw images are not considered as holes.
if qemu_nbd.version() >= (6, 0, 0):
    FMT_HOLE_FLAGS_PARAMS[0][1] |= nbd.STATE_HOLE


@pytest.mark.parametrize("fmt,hole_flags", FMT_HOLE_FLAGS_PARAMS)
def test_base_allocation_empty(nbd_server, user_file, fmt, hole_flags):
    size = 1024**3
    create_image(user_file.path, fmt, size)

    nbd_server.image = user_file.path
    nbd_server.fmt = fmt
    nbd_server.start()

    with nbd.open(nbd_server.url) as c:
        # Entire image.
        extents = list(nbdutil.extents(c))
        assert extents == [nbd.Extent(length=size, flags=hole_flags)]

        # First block.
        extents = list(nbdutil.extents(c, length=4096))
        assert extents == [nbd.Extent(length=4096, flags=hole_flags)]

        # Last block.
        extents = list(nbdutil.extents(c, offset=size-4096, length=4096))
        assert extents == [nbd.Extent(length=4096, flags=hole_flags)]

        # Some block.
        extents = list(nbdutil.extents(c, offset=4096, length=4096))
        assert extents == [nbd.Extent(length=4096, flags=hole_flags)]

        # Unaligned start.
        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 1))
        assert extents == [nbd.Extent(length=4096 + 1, flags=hole_flags)]

        # Unaligned end.
        extents = list(nbdutil.extents(c, offset=4096, length=4096 + 1))
        assert extents == [nbd.Extent(length=4096 + 1, flags=hole_flags)]

        # Unaligned start and end.
        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 2))
        assert extents == [nbd.Extent(length=4096 + 2, flags=hole_flags)]


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_base_allocation_full(nbd_server, user_file, fmt):
    size = 1024**2
    create_image(user_file.path, fmt, size)

    nbd_server.image = user_file.path
    nbd_server.fmt = fmt
    nbd_server.start()

    with nbd.open(nbd_server.url) as c:
        c.write(0, b"x" * size)

        # Entire image.
        extents = list(nbdutil.extents(c))
        assert extents == [nbd.Extent(length=size, flags=0)]

        # First block.
        extents = list(nbdutil.extents(c, length=4096))
        assert extents == [nbd.Extent(length=4096, flags=0)]

        # Last block.
        extents = list(nbdutil.extents(c, offset=size-4096, length=4096))
        assert extents == [nbd.Extent(length=4096, flags=0)]

        # Some block.
        extents = list(nbdutil.extents(c, offset=4096, length=4096))
        assert extents == [nbd.Extent(length=4096, flags=0)]

        # Unaligned start.
        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 1))
        assert extents == [nbd.Extent(length=4096 + 1, flags=0)]

        # Unaligned end.
        extents = list(nbdutil.extents(c, offset=4096, length=4096 + 1))
        assert extents == [nbd.Extent(length=4096 + 1, flags=0)]

        # Unaligned start and end.
        extents = list(nbdutil.extents(c, offset=4096 - 1, length=4096 + 2))
        assert extents == [nbd.Extent(length=4096 + 2, flags=0)]


@pytest.mark.parametrize("fmt,hole_flags", FMT_HOLE_FLAGS_PARAMS)
def test_base_allocation_some_data(nbd_server, user_file, fmt, hole_flags):
    size = 1024**3
    create_image(user_file.path, fmt, size)

    nbd_server.image = user_file.path
    nbd_server.fmt = fmt
    nbd_server.start()

    # Use qcow2 cluster size to avoid inconsistent results on CentOS and
    # Fedora.
    data_length = 64 * 1024
    zero_length = size // 2 - data_length

    with nbd.open(nbd_server.url) as c:
        # Create 4 extents: data, zero, data, zero.
        c.write(0, b"x" * data_length)
        c.write(size // 2, b"x" * data_length)

        extents = list(nbdutil.extents(c))

    assert extents == [
        nbd.Extent(length=data_length, flags=0),
        nbd.Extent(length=zero_length, flags=hole_flags),
        nbd.Extent(length=data_length, flags=0),
        nbd.Extent(length=zero_length, flags=hole_flags),
    ]


@pytest.mark.parametrize("fmt,hole_flags", FMT_HOLE_FLAGS_PARAMS)
def test_base_allocation_some_data_unaligned(
        nbd_server, user_file, fmt, hole_flags):
    size = 1024**2
    create_image(user_file.path, fmt, size)

    nbd_server.image = user_file.path
    nbd_server.fmt = fmt
    nbd_server.start()

    data_length = 64 * 1024
    data_offset = 2 * data_length

    with nbd.open(nbd_server.url) as c:
        # Create 3 extents: zero, data, zero.
        c.write(data_offset, b"x" * data_length)

        # Unaligned part from first extent and last extent.
        extents = list(nbdutil.extents(c, data_offset - 1, data_length + 2))
        assert extents == [
            nbd.Extent(length=1, flags=hole_flags),
            nbd.Extent(length=data_length, flags=0),
            nbd.Extent(length=1, flags=hole_flags),
        ]

        # Unaligned part from second extent.
        extents = list(nbdutil.extents(c, data_offset + 1, data_length - 2))
        assert extents == [
            nbd.Extent(length=data_length - 2, flags=0),
        ]

        # Unaligned part from second and last extents.
        extents = list(nbdutil.extents(c, data_offset + 1, data_length))
        assert extents == [
            nbd.Extent(length=data_length - 1, flags=0),
            nbd.Extent(length=1, flags=hole_flags),
        ]


def test_base_allocation_many_extents(nbd_server, user_file):
    # Tested only with raw since qcow2 minimal extent is cluster size (64K),
    # and writing 1000 extents (62.5 MiB) will be too slow in the CI.

    # Extents must be multiple of file system block size.
    extent_length = os.statvfs(user_file.path).f_bsize

    # Use number which is not a multiple of our buffer capacity (1024 extents)
    # to ensure we read partial buffers correctly.
    extents_count = 2000

    size = extents_count * extent_length
    create_image(user_file.path, "raw", size)

    nbd_server.image = user_file.path
    nbd_server.fmt = "raw"
    nbd_server.start()

    with nbd.open(nbd_server.url) as c:
        # Write data to all even extents.
        data = b"x" * extent_length
        for i in range(0, size, extent_length * 2):
            c.write(i, data)

        extents = list(nbdutil.extents(c))

    assert len(extents) == extents_count

    for i, ext in enumerate(extents):
        assert ext.length == extent_length
        assert ext.zero == bool(i % 2)


def test_extents_reply_error(nbd_server, user_file):
    """
    The server SHOULD return NBD_EINVAL if it receives a NBD_CMD_BLOCK_STATUS
    request including one or more sectors beyond the size of the device.
    """
    size = 1024**2
    create_image(user_file.path, "raw", size)

    nbd_server.image = user_file.path
    nbd_server.fmt = "raw"
    nbd_server.start()

    with nbd.open(nbd_server.url) as c:
        with pytest.raises(nbd.ReplyError) as e:
            c.extents(0, size + 1)

        # Server should return this, qemu does.
        assert e.value.code == errno.EINVAL

        # The next request should succeed.
        assert c.read(4096, 1) == b"\0"


# Communicate with qemu builtin NBD server


@pytest.mark.parametrize("fmt", ["raw", "qcow2"])
def test_full_backup_handshake(tmpdir, fmt, nbd_sock):
    image = str(tmpdir.join("image"))
    create_image(image, fmt, 1024**3)

    with backup.full_backup(tmpdir, image, fmt, nbd_sock):
        with nbd.Client(nbd_sock, "sda") as c:
            # TODO: test transmission_flags?
            assert c.export_size == 1024**3
            assert c.minimum_block_size == 1
            assert c.preferred_block_size == 4096
            assert c.maximum_block_size == 32 * 1024**2
            # Both available in in current qemu (>= 5.2.0).
            assert c.has_base_allocation
            assert c.has_allocation_depth


@pytest.mark.parametrize("fmt", [
    pytest.param(
        "raw",
        marks=pytest.mark.xfail(
            distro.is_centos() and ci.is_ovirt(),
            reason="unaligned write fails on el8/oVirt CI")
    ),
    pytest.param("qcow2"),
])
def test_full_backup_single_image(tmpdir, user_file, fmt, nbd_sock):
    chunk_size = 1024**3
    disk_size = 5 * chunk_size

    # Create disk
    create_image(user_file.path, fmt, disk_size)

    # Pupulate disk with data.
    with qemu_nbd.open(user_file.path, fmt) as d:
        for offset in range(0, disk_size, chunk_size):
            d.write(offset, b"%d\n" % offset)
        d.flush()

    checkpoint = "check1" if fmt == "qcow2" else None

    # Start full backup and copy the data, veifying what we read.
    with backup.full_backup(
            tmpdir, user_file.path, fmt, nbd_sock, checkpoint=checkpoint):
        verify_full_backup(nbd_sock, "sda")

    if checkpoint:
        bitmaps = list_bitmaps(user_file.path)
        assert len(bitmaps) == 1
        assert bitmaps[0]["name"] == checkpoint


@pytest.mark.parametrize("checkpoint", [
    pytest.param(None, id="no-checkpoint"),
    pytest.param("check1", id="with-checkpoint"),
])
def test_full_backup_complete_chain(tmpdir, nbd_sock, checkpoint):
    depth = 3
    chunk_size = 1024**2
    disk_size = depth * chunk_size

    for i in range(depth):
        # Create disk based on previous one.
        disk = str(tmpdir.join("disk.{}".format(i)))
        if i == 0:
            qemu_img.create(disk, "qcow2", size=disk_size)
        else:
            qemu_img.create(
                disk,
                "qcow2",
                backing_file="disk.{}".format(i - 1),
                backing_format="qcow2")

        # This data can be read only from this disk.
        with qemu_nbd.open(disk, "qcow2") as d:
            offset = i * chunk_size
            d.write(offset, b"%d\n" % offset)
            d.flush()

    # Start full backup and copy the data, veifying what we read.
    with backup.full_backup(
            tmpdir, disk, "qcow2", nbd_sock, checkpoint=checkpoint):
        verify_full_backup(nbd_sock, "sda")

    if checkpoint:
        bitmaps = list_bitmaps(disk)
        assert len(bitmaps) == 1
        assert bitmaps[0]["name"] == checkpoint


def list_bitmaps(image):
    info = qemu_img.info(image)
    return info["format-specific"]["data"]["bitmaps"]


def verify_full_backup(sock, export_name):
    with nbd.Client(sock, export_name) as c:
        log.debug("Backing up data with nbd client")
        offset = 0
        for ext in nbdutil.extents(c):
            if not ext.zero:
                expected = b"%d\n" % offset
                data = c.read(offset, len(expected))
                assert data == expected
            offset += ext.length


def test_extent_base_allocation():
    # Allocated area with data.
    data = nbd.Extent.pack(4096, 0)
    ext = nbd.Extent.unpack(data)
    assert not ext.zero
    assert not ext.hole
    assert ext.flags == 0

    # Allocated area that reads as zero.
    data = nbd.Extent.pack(4096, nbd.STATE_ZERO)
    ext = nbd.Extent.unpack(data)
    assert ext.zero
    assert not ext.hole
    assert ext.flags == nbd.STATE_ZERO

    # Unallocated area that reads as zero.
    data = nbd.Extent.pack(4096, nbd.STATE_ZERO | nbd.STATE_HOLE)
    ext = nbd.Extent.unpack(data)
    assert ext.zero
    # Hole can be detected only when using depth data with qcow2 image.
    assert not ext.hole
    assert ext.flags == nbd.STATE_ZERO | nbd.STATE_HOLE

    # Ignore unexpected bits
    data = nbd.Extent.pack(4096, 0xffff)
    ext = nbd.Extent.unpack(data)
    assert ext.zero
    assert not ext.hole
    assert not ext.dirty
    assert ext.flags == nbd.STATE_HOLE | nbd.STATE_ZERO


def test_extent_dirty_bitmap():
    # Clean area.
    data = nbd.Extent.pack(4096, 0)
    ext = nbd.Extent.unpack(data, nbd.Extent.DIRTY)
    assert not ext.dirty
    assert ext.flags == 0

    # Dirty area.
    data = nbd.Extent.pack(4096, nbd.STATE_DIRTY)
    ext = nbd.Extent.unpack(data, nbd.Extent.DIRTY)
    assert ext.dirty
    # nbd.STATE_DIRTY is mapped to nbd.EXTENT_DIRTY to allow merging allocation
    # and dirty bitmap info in same extent.
    assert ext.flags == nbd.EXTENT_DIRTY

    # Ignore unexpected bits
    data = nbd.Extent.pack(4096, 0xffff)
    ext = nbd.Extent.unpack(data, nbd.Extent.DIRTY)
    assert not ext.zero
    assert not ext.hole
    assert ext.dirty
    assert ext.flags == nbd.EXTENT_DIRTY


def test_extent_depth():
    # Unallocated extent.
    depth_data = nbd.Extent.pack(65536, 0)
    ext = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)
    assert ext.length == 65536
    assert not ext.zero
    # Hole status is determined only by the EXTENT_BACKING bit.
    assert ext.hole
    assert not ext.dirty
    assert ext.flags == nbd.EXTENT_BACKING

    # Allocated extents from top layer.
    data = nbd.Extent.pack(4096, 1)
    ext = nbd.Extent.unpack(data, nbd.Extent.DEPTH)
    assert not ext.zero
    assert not ext.hole
    assert not ext.dirty
    assert ext.flags == 0

    # Allocated extents from top backing file.
    data = nbd.Extent.pack(4096, 2)
    ext = nbd.Extent.unpack(data, nbd.Extent.DEPTH)
    assert not ext.zero
    assert not ext.hole
    assert not ext.dirty
    assert ext.flags == 0


def test_extent_compare():
    assert nbd.Extent(4096, 0) == nbd.Extent(4096, 0)
    assert nbd.Extent(4096, 0) != nbd.Extent(4096, nbd.STATE_ZERO)


def test_extent_merge_clean_hole():
    alloc_data = nbd.Extent.pack(4096, nbd.STATE_ZERO | nbd.STATE_HOLE)
    alloc = nbd.Extent.unpack(alloc_data)

    dirty_data = nbd.Extent.pack(4096, 0)
    dirty = nbd.Extent.unpack(dirty_data, nbd.Extent.DIRTY)

    merged = nbd.Extent(4096, alloc.flags | dirty.flags)

    assert merged.length == 4096
    assert merged.zero is True
    # Hole can be detected only when using depth data with qcow2 image.
    assert merged.hole is False
    assert not merged.dirty


def test_extent_merge_dirty_data():
    alloc_data = nbd.Extent.pack(4096, 0)
    alloc = nbd.Extent.unpack(alloc_data, nbd.Extent.DIRTY)

    dirty_data = nbd.Extent.pack(4096, nbd.STATE_DIRTY)
    dirty = nbd.Extent.unpack(dirty_data, nbd.Extent.DIRTY)

    merged = nbd.Extent(4096, alloc.flags | dirty.flags)

    assert merged.length == 4096
    assert not merged.zero
    assert not merged.hole
    assert merged.dirty


def test_extent_merge_qcow2_unallocated_cluster():
    alloc_data = nbd.Extent.pack(65536, nbd.STATE_ZERO | nbd.STATE_HOLE)
    alloc = nbd.Extent.unpack(alloc_data)

    depth_data = nbd.Extent.pack(65536, 0)
    depth = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)

    merged = nbd.Extent(65536, alloc.flags | depth.flags)

    assert merged.length == 65536
    assert merged.zero is True
    assert merged.hole is True


def test_extent_merge_qcow2_zero_cluster():
    alloc_data = nbd.Extent.pack(65536, nbd.STATE_ZERO | nbd.STATE_HOLE)
    alloc = nbd.Extent.unpack(alloc_data)

    depth_data = nbd.Extent.pack(65536, 1)
    depth = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)

    merged = nbd.Extent(65536, alloc.flags | depth.flags)

    assert merged.length == 65536
    assert merged.zero is True
    assert merged.hole is False


def test_extent_merge_qcow2_data_cluster():
    alloc_data = nbd.Extent.pack(65536, 0)
    alloc = nbd.Extent.unpack(alloc_data)

    depth_data = nbd.Extent.pack(65536, 1)
    depth = nbd.Extent.unpack(depth_data, nbd.Extent.DEPTH)

    merged = nbd.Extent(65536, alloc.flags | depth.flags)

    assert merged.length == 65536
    assert merged.zero is False
    assert merged.hole is False


def test_reply_error_structured():
    s = str(nbd.ReplyError(28, "writing to file failed"))
    assert s == "Writing to file failed: [Error 28] No space left on device"


def test_reply_error_simple():
    s = str(nbd.ReplyError(1, "Simple reply failed"))
    assert s == "Simple reply failed: [Error 1] Operation not permitted"


def test_reply_error_unknown_code():
    s = str(nbd.ReplyError(-1, "bad error"))
    assert s == "Bad error: [Error -1] Unknown error -1"


def test_reply_error_empty_message():
    s = str(nbd.ReplyError(5, ""))
    assert s == "Server error: [Error 5] Input/output error"


def create_image(path, fmt, size):
    if fmt == "raw":
        # qemu-img allocates the first block on Fedora, but not on CentOS 8.0.
        # Allocate manually for consistent results.
        # TODO: Use qemu-img when we have CentOS 8.1 AV.
        with io.open(path, "wb") as f:
            f.truncate(size)
    else:
        qemu_img.create(path, "qcow2", size=size)



# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
userstorage configuration
"""

from userstorage import File, Mount, LoopDevice

GiB = 1024**3

BASE_DIR = "/var/tmp/imageio-storage"

BACKENDS = {

    # This storage is generally not interesting for testing, but since it does
    # not support fallocate(), it is a good way to test the fallback to manualy
    # writing zeroes.
    "file-512-ext2": File(
        Mount(
            LoopDevice(
                base_dir=BASE_DIR,
                name="file-512-ext2",
                size=GiB,
                sector_size=512,
            ),
            fstype="ext2",
        )
    ),

    "file-512-ext4": File(
        Mount(
            LoopDevice(
                base_dir=BASE_DIR,
                name="file-512-ext4",
                size=GiB,
                sector_size=512,
            ),
            fstype="ext4",
        )
    ),

    "file-512-xfs": File(
        Mount(
            LoopDevice(
                base_dir=BASE_DIR,
                name="file-512-xfs",
                size=GiB,
                sector_size=512,
                # Fails to mount on Jenkins when running on slave with kernel
                # 3.10. Let's make it optional so we can test other storage.
                required=False,
            ),
            fstype="xfs",
        )
    ),

    "file-4k-ext2": File(
        Mount(
            LoopDevice(
                base_dir=BASE_DIR,
                name="file-4k-ext2",
                size=GiB,
                sector_size=4096,
                required=False,
            ),
            fstype="ext2",
        )
    ),

    "file-4k-ext4": File(
        Mount(
            LoopDevice(
                base_dir=BASE_DIR,
                name="file-4k-ext4",
                size=GiB,
                sector_size=4096,
                required=False,
            ),
            fstype="ext4",
        )
    ),

    "file-4k-xfs": File(
        Mount(
            LoopDevice(
                base_dir=BASE_DIR,
                name="file-4k-xfs",
                size=GiB,
                sector_size=4096,
                required=False,
            ),
            fstype="xfs",
        )
    ),

    "block-512": LoopDevice(
        base_dir=BASE_DIR,
        name="block-512",
        size=GiB,
        sector_size=512,
        required=False,
    ),

    "block-4k": LoopDevice(
        base_dir=BASE_DIR,
        name="block-4k",
        size=GiB,
        sector_size=4096,
        required=False,
    ),

}
